from typing import Any, Dict, List, Union
from instana.log import logger


def parse_service_pair(pair: str) -> List[str]:
    """
    Parses a pair string to prepare a list of ignored endpoints.

    @param pair: String format:
        - "service1:endpoint1,endpoint2" or "service1:endpoint1" or "service1"
    @return: List of strings in format ["service1.endpoint1", "service1.endpoint2", "service2"]
    """
    pair_list = []
    if ":" in pair:
        service, endpoints = pair.split(":", 1)
        service = service.strip()
        endpoint_list = [ep.strip() for ep in endpoints.split(",") if ep.strip()]

        for endpoint in endpoint_list:
            pair_list.append(f"{service}.{endpoint}")
    else:
        pair_list.append(pair)
    return pair_list


def parse_ignored_endpoints_string(params: str) -> List[str]:
    """
    Parses a string to prepare a list of ignored endpoints.

    @param params: String format:
        - "service1:endpoint1,endpoint2;service2:endpoint3" or "service1;service2"
    @return: List of strings in format ["service1.endpoint1", "service1.endpoint2", "service2"]
    """
    ignore_endpoints = []
    if params:
        service_pairs = params.lower().split(";")

        for pair in service_pairs:
            if pair.strip():
                ignore_endpoints += parse_service_pair(pair)
    return ignore_endpoints


def parse_ignored_endpoints_dict(params: Dict[str, Any]) -> List[str]:
    """
    Parses a dictionary to prepare a list of ignored endpoints.

    @param params: Dict format:
        - {"service1": ["endpoint1", "endpoint2"], "service2": ["endpoint3"]}
    @return: List of strings in format ["service1.endpoint1", "service1.endpoint2", "service2"]
    """
    ignore_endpoints = []

    for service, endpoints in params.items():
        if not endpoints:  # filtering all service
            ignore_endpoints.append(service.lower())
        else:  # filtering specific endpoints
            for endpoint in endpoints:
                ignore_endpoints.append(f"{service.lower()}.{endpoint.lower()}")

    return ignore_endpoints


def parse_ignored_endpoints(params: Union[Dict[str, Any], str]) -> List[str]:
    """
    Parses input to prepare a list for ignored endpoints.

    @param params: Can be either:
        - String: "service1:endpoint1,endpoint2;service2:endpoint3" or "service1;service2"
        - Dict: {"service1": ["endpoint1", "endpoint2"], "service2": ["endpoint3"]}
    @return: List of strings in format ["service1.endpoint1", "service1.endpoint2", "service2"]
    """
    try:
        if isinstance(params, str):
            return parse_ignored_endpoints_string(params)
        elif isinstance(params, dict):
            return parse_ignored_endpoints_dict(params)
        else:
            return []
    except Exception as e:
        logger.debug("Error parsing ignored endpoints: %s", str(e))
        return []



# (c) Copyright IBM Corp. 2025
from typing import TYPE_CHECKING, Any, Dict, Optional

from opentelemetry.trace.span import format_span_id

from instana.log import logger
from instana.propagators.base_propagator import BasePropagator, CarrierT
from instana.util.ids import hex_id_limited

if TYPE_CHECKING:
    from instana.span_context import SpanContext


class KafkaPropagator(BasePropagator):
    """
    Instana Propagator for Format.KAFKA_HEADERS.

    The KAFKA_HEADERS format deals with key-values with string to string mapping.
    The character set should be restricted to HTTP compatible.
    """

    def __init__(self) -> None:
        super(KafkaPropagator, self).__init__()

    # Assisted by watsonx Code Assistant
    def extract_carrier_headers(self, carrier: CarrierT) -> Dict[str, Any]:
        """
        Extracts headers from a carrier object.

        Args:
            carrier (CarrierT): The carrier object to extract headers from.

        Returns:
            Dict[str, Any]: A dictionary containing the extracted headers.
        """
        dc = {}
        try:
            if isinstance(carrier, list):
                for header in carrier:
                    if isinstance(header, tuple):
                        dc[header[0]] = header[1]
                    elif isinstance(header, dict):
                        for k, v in header.items():
                            dc[k] = v
            else:
                dc = self.extract_headers_dict(carrier)
        except Exception:
            logger.debug(
                f"kafka_propagator extract_headers_list: Couldn't convert - {carrier}"
            )

        return dc

    def extract(
        self, carrier: CarrierT, disable_w3c_trace_context: bool = False
    ) -> Optional["SpanContext"]:
        """
        This method overrides one of the Base classes as with the introduction
        of W3C trace context for the Kafka requests more extracting steps and
        logic was required.

        Args:
            carrier (CarrierT): The carrier object to extract headers from.
            disable_w3c_trace_context (bool): A flag to disable the W3C trace context.

        Returns:
            Optional["SpanContext"]: The extracted span context or None.
        """
        try:
            headers = self.extract_carrier_headers(carrier=carrier)
            return super(KafkaPropagator, self).extract(
                carrier=headers,
                disable_w3c_trace_context=disable_w3c_trace_context,
            )

        except Exception:
            logger.debug("kafka_propagator extract error:", exc_info=True)

    # Assisted by watsonx Code Assistant
    def inject(
        self,
        span_context: "SpanContext",
        carrier: CarrierT,
        disable_w3c_trace_context: bool = True,
    ) -> None:
        """
        Inject the trace context into a carrier.

        Args:
            span_context (SpanContext): The SpanContext object containing trace information.
            carrier (CarrierT): The carrier object to store the trace context.
            disable_w3c_trace_context (bool, optional): A boolean flag to disable W3C trace context. Defaults to True.

        Returns:
            None
        """
        trace_id = span_context.trace_id
        span_id = span_context.span_id
        dictionary_carrier = self.extract_carrier_headers(carrier)

        if dictionary_carrier:
            # Suppression `level` made in the child context or in the parent context
            # has priority over any non-suppressed `level` setting
            child_level = int(
                self.extract_instana_headers(dictionary_carrier)[2] or "1"
            )
            span_context.level = min(child_level, span_context.level)

        serializable_level = str(span_context.level)

        def inject_key_value(carrier, key, value):
            if isinstance(carrier, list):
                carrier.append((key, value))
            elif isinstance(carrier, dict) or "__setitem__" in dir(carrier):
                carrier[key] = value
            else:
                raise Exception(
                    f"KafkaPropagator: Unsupported carrier type {type(carrier)}",
                )

        try:
            inject_key_value(
                carrier,
                self.KAFKA_HEADER_KEY_L_S,
                serializable_level.encode("utf-8"),
            )
            inject_key_value(
                carrier,
                self.KAFKA_HEADER_KEY_T,
                hex_id_limited(trace_id).encode("utf-8"),
            )
            inject_key_value(
                carrier,
                self.KAFKA_HEADER_KEY_S,
                format_span_id(span_id).encode("utf-8"),
            )

        except Exception:
            logger.debug("KafkaPropagator - inject error:", exc_info=True)



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

import os
import pytest
from typing import Generator
from moto import mock_aws
import boto3

from instana.singletons import tracer, agent
from tests.helpers import get_first_span_by_filter

pwd = os.path.dirname(os.path.abspath(__file__))
upload_filename = os.path.abspath(pwd + "/../../data/boto3/test_upload_file.jpg")
download_target_filename = os.path.abspath(
    pwd + "/../../data/boto3/download_target_file.asdf"
)


class TestS3:
    @pytest.fixture(autouse=True)
    def _resource(self) -> Generator[None, None, None]:
        """Setup and Teardown"""
        # Clear all spans before a test run
        self.recorder = tracer.span_processor
        self.recorder.clear_spans()
        self.mock = mock_aws()
        self.mock.start()
        self.s3 = boto3.client("s3", region_name="us-east-1")
        yield
        # Stop Moto after each test
        self.mock.stop()
        agent.options.allow_exit_as_root = False

    def test_vanilla_create_bucket(self) -> None:
        self.s3.create_bucket(Bucket="aws_bucket_name")

        result = self.s3.list_buckets()
        assert len(result["Buckets"]) == 1
        assert result["Buckets"][0]["Name"] == "aws_bucket_name"

    def test_s3_create_bucket(self) -> None:
        with tracer.start_as_current_span("test"):
            self.s3.create_bucket(Bucket="aws_bucket_name")

        result = self.s3.list_buckets()
        assert len(result["Buckets"]) == 1
        assert result["Buckets"][0]["Name"] == "aws_bucket_name"

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        filter = lambda span: span.n == "sdk"  # noqa: E731
        test_span = get_first_span_by_filter(spans, filter)
        assert test_span

        filter = lambda span: span.n == "s3"  # noqa: E731
        s3_span = get_first_span_by_filter(spans, filter)
        assert s3_span

        assert s3_span.t == test_span.t
        assert s3_span.p == test_span.s

        assert not test_span.ec
        assert not s3_span.ec

        assert s3_span.data["s3"]["op"] == "CreateBucket"
        assert s3_span.data["s3"]["bucket"] == "aws_bucket_name"

    def test_s3_create_bucket_as_root_exit_span(self) -> None:
        agent.options.allow_exit_as_root = True
        self.s3.create_bucket(Bucket="aws_bucket_name")

        agent.options.allow_exit_as_root = False
        self.s3.list_buckets()

        spans = self.recorder.queued_spans()
        assert len(spans) == 1

        s3_span = spans[0]
        assert s3_span

        assert not s3_span.ec

        assert s3_span.data["s3"]["op"] == "CreateBucket"
        assert s3_span.data["s3"]["bucket"] == "aws_bucket_name"

    def test_s3_list_buckets(self) -> None:
        with tracer.start_as_current_span("test"):
            result = self.s3.list_buckets()

        assert len(result["Buckets"]) == 0
        assert result["ResponseMetadata"]["HTTPStatusCode"] == 200

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        filter = lambda span: span.n == "sdk"  # noqa: E731
        test_span = get_first_span_by_filter(spans, filter)
        assert test_span

        filter = lambda span: span.n == "s3"  # noqa: E731
        s3_span = get_first_span_by_filter(spans, filter)
        assert s3_span

        assert s3_span.t == test_span.t
        assert s3_span.p == test_span.s

        assert not test_span.ec
        assert not s3_span.ec

        assert s3_span.data["s3"]["op"] == "ListBuckets"
        assert not s3_span.data["s3"]["bucket"]

    def test_s3_vanilla_upload_file(self) -> None:
        object_name = "aws_key_name"
        bucket_name = "aws_bucket_name"

        self.s3.create_bucket(Bucket=bucket_name)
        result = self.s3.upload_file(upload_filename, bucket_name, object_name)
        assert not result

    def test_s3_upload_file(self) -> None:
        object_name = "aws_key_name"
        bucket_name = "aws_bucket_name"

        self.s3.create_bucket(Bucket=bucket_name)

        with tracer.start_as_current_span("test"):
            self.s3.upload_file(upload_filename, bucket_name, object_name)

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        filter = lambda span: span.n == "sdk"  # noqa: E731
        test_span = get_first_span_by_filter(spans, filter)
        assert test_span

        filter = lambda span: span.n == "s3"  # noqa: E731
        s3_span = get_first_span_by_filter(spans, filter)
        assert s3_span

        assert s3_span.t == test_span.t
        assert s3_span.p == test_span.s

        assert not test_span.ec
        assert not s3_span.ec

        assert s3_span.data["s3"]["op"] == "UploadFile"
        assert s3_span.data["s3"]["bucket"] == "aws_bucket_name"

    def test_s3_upload_file_obj(self) -> None:
        object_name = "aws_key_name"
        bucket_name = "aws_bucket_name"

        self.s3.create_bucket(Bucket=bucket_name)

        with tracer.start_as_current_span("test"):
            with open(upload_filename, "rb") as fd:
                self.s3.upload_fileobj(fd, bucket_name, object_name)

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        filter = lambda span: span.n == "sdk"  # noqa: E731
        test_span = get_first_span_by_filter(spans, filter)
        assert test_span

        filter = lambda span: span.n == "s3"  # noqa: E731
        s3_span = get_first_span_by_filter(spans, filter)
        assert s3_span

        assert s3_span.t == test_span.t
        assert s3_span.p == test_span.s

        assert not test_span.ec
        assert not s3_span.ec

        assert s3_span.data["s3"]["op"] == "UploadFileObj"
        assert s3_span.data["s3"]["bucket"] == "aws_bucket_name"

    def test_s3_download_file(self) -> None:
        object_name = "aws_key_name"
        bucket_name = "aws_bucket_name"

        self.s3.create_bucket(Bucket=bucket_name)
        self.s3.upload_file(upload_filename, bucket_name, object_name)

        with tracer.start_as_current_span("test"):
            self.s3.download_file(bucket_name, object_name, download_target_filename)

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        filter = lambda span: span.n == "sdk"  # noqa: E731
        test_span = get_first_span_by_filter(spans, filter)
        assert test_span

        filter = lambda span: span.n == "s3"  # noqa: E731
        s3_span = get_first_span_by_filter(spans, filter)
        assert s3_span

        assert s3_span.t == test_span.t
        assert s3_span.p == test_span.s

        assert not test_span.ec
        assert not s3_span.ec

        assert s3_span.data["s3"]["op"] == "DownloadFile"
        assert s3_span.data["s3"]["bucket"] == "aws_bucket_name"

    def test_s3_download_file_obj(self) -> None:
        object_name = "aws_key_name"
        bucket_name = "aws_bucket_name"

        self.s3.create_bucket(Bucket=bucket_name)
        self.s3.upload_file(upload_filename, bucket_name, object_name)

        with tracer.start_as_current_span("test"):
            with open(download_target_filename, "wb") as fd:
                self.s3.download_fileobj(bucket_name, object_name, fd)

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        filter = lambda span: span.n == "sdk"  # noqa: E731
        test_span = get_first_span_by_filter(spans, filter)
        assert test_span

        filter = lambda span: span.n == "s3"  # noqa: E731
        s3_span = get_first_span_by_filter(spans, filter)
        assert s3_span

        assert s3_span.t == test_span.t
        assert s3_span.p == test_span.s

        assert not test_span.ec
        assert not s3_span.ec

        assert s3_span.data["s3"]["op"] == "DownloadFileObj"
        assert s3_span.data["s3"]["bucket"] == "aws_bucket_name"

    def test_s3_list_obj(self) -> None:
        bucket_name = "aws_bucket_name"

        self.s3.create_bucket(Bucket=bucket_name)

        with tracer.start_as_current_span("test"):
            self.s3.list_objects(Bucket=bucket_name)

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        filter = lambda span: span.n == "sdk"  # noqa: E731
        test_span = get_first_span_by_filter(spans, filter)
        assert test_span

        filter = lambda span: span.n == "s3"  # noqa: E731
        s3_span = get_first_span_by_filter(spans, filter)
        assert s3_span

        assert s3_span.t == test_span.t
        assert s3_span.p == test_span.s

        assert not test_span.ec
        assert not s3_span.ec

        assert s3_span.data["s3"]["op"] == "ListObjects"
        assert s3_span.data["s3"]["bucket"] == "aws_bucket_name"



from typing import Generator

import pytest

from instana.util.config import (
    parse_ignored_endpoints,
    parse_ignored_endpoints_dict,
    parse_service_pair,
)


class TestConfig:
    @pytest.fixture(autouse=True)
    def _resource(self) -> Generator[None, None, None]:
        yield

    def test_parse_service_pair(self) -> None:
        test_string = "service1:endpoint1,endpoint2"
        response = parse_service_pair(test_string)
        assert response == ["service1.endpoint1", "service1.endpoint2"]

        test_string = "service1;service2"
        response = parse_ignored_endpoints(test_string)
        assert response == ["service1", "service2"]

        test_string = "service1"
        response = parse_ignored_endpoints(test_string)
        assert response == ["service1"]

        test_string = ";"
        response = parse_ignored_endpoints(test_string)
        assert response == []

        test_string = "service1:endpoint1,endpoint2;;;service2:endpoint1;;"
        response = parse_ignored_endpoints(test_string)
        assert response == [
            "service1.endpoint1",
            "service1.endpoint2",
            "service2.endpoint1",
        ]

        test_string = ""
        response = parse_ignored_endpoints(test_string)
        assert response == []

    def test_parse_ignored_endpoints_string(self) -> None:
        test_string = "service1:endpoint1,endpoint2"
        response = parse_service_pair(test_string)
        assert response == ["service1.endpoint1", "service1.endpoint2"]

        test_string = "service1;service2"
        response = parse_ignored_endpoints(test_string)
        assert response == ["service1", "service2"]

        test_string = "service1"
        response = parse_ignored_endpoints(test_string)
        assert response == ["service1"]

        test_string = ";"
        response = parse_ignored_endpoints(test_string)
        assert response == []

        test_string = "service1:endpoint1,endpoint2;;;service2:endpoint1;;"
        response = parse_ignored_endpoints(test_string)
        assert response == [
            "service1.endpoint1",
            "service1.endpoint2",
            "service2.endpoint1",
        ]

        test_string = ""
        response = parse_ignored_endpoints(test_string)
        assert response == []

    def test_parse_ignored_endpoints_dict(self) -> None:
        test_dict = {"service1": ["endpoint1", "endpoint2"]}
        response = parse_ignored_endpoints_dict(test_dict)
        assert response == ["service1.endpoint1", "service1.endpoint2"]

        test_dict = {"SERVICE1": ["ENDPOINT1", "ENDPOINT2"]}
        response = parse_ignored_endpoints_dict(test_dict)
        assert response == ["service1.endpoint1", "service1.endpoint2"]

        test_dict = {"service1": [], "service2": []}
        response = parse_ignored_endpoints_dict(test_dict)
        assert response == ["service1", "service2"]

        test_dict = {"service1": []}
        response = parse_ignored_endpoints_dict(test_dict)
        assert response == ["service1"]

        test_dict = {}
        response = parse_ignored_endpoints_dict(test_dict)
        assert response == []

    def test_parse_ignored_endpoints(self) -> None:
        test_pair = "service1:endpoint1,endpoint2"
        response = parse_ignored_endpoints(test_pair)
        assert response == ["service1.endpoint1", "service1.endpoint2"]

        test_pair = "service1;service2"
        response = parse_ignored_endpoints(test_pair)
        assert response == ["service1", "service2"]

        test_pair = "service1"
        response = parse_ignored_endpoints(test_pair)
        assert response == ["service1"]

        test_pair = ";"
        response = parse_ignored_endpoints(test_pair)
        assert response == []

        test_pair = "service1:endpoint1,endpoint2;;;service2:endpoint1;;"
        response = parse_ignored_endpoints(test_pair)
        assert response == [
            "service1.endpoint1",
            "service1.endpoint2",
            "service2.endpoint1",
        ]

        test_pair = ""
        response = parse_ignored_endpoints(test_pair)
        assert response == []

        test_dict = {"service1": ["endpoint1", "endpoint2"]}
        response = parse_ignored_endpoints(test_dict)
        assert response == ["service1.endpoint1", "service1.endpoint2"]

        test_dict = {"service1": [], "service2": []}
        response = parse_ignored_endpoints(test_dict)
        assert response == ["service1", "service2"]

        test_dict = {"service1": []}
        response = parse_ignored_endpoints(test_dict)
        assert response == ["service1"]

        test_dict = {}
        response = parse_ignored_endpoints(test_dict)
        assert response == []



# (c) Copyright IBM Corp. 2025

try:
    import aioamqp
    from typing import Any, Callable, Dict, Tuple

    import wrapt
    from opentelemetry.trace.status import StatusCode

    from instana.log import logger
    from instana.util.traceutils import get_tracer_tuple, tracing_is_off

    @wrapt.patch_function_wrapper("aioamqp.channel", "Channel.basic_publish")
    async def basic_publish_with_instana(
        wrapped: Callable[..., aioamqp.connect],
        instance: object,
        argv: Tuple[object, Tuple[object, ...]],
        kwargs: Dict[str, Any],
    ) -> object:
        if tracing_is_off():
            return await wrapped(*argv, **kwargs)

        tracer, parent_span, _ = get_tracer_tuple()
        parent_context = parent_span.get_span_context() if parent_span else None
        with tracer.start_as_current_span(
            "aioamqp-publisher", span_context=parent_context
        ) as span:
            try:
                span.set_attribute("aioamqp.exchange", argv[0])
                return await wrapped(*argv, **kwargs)
            except Exception as exc:
                span.record_exception(exc)
                logger.debug(f"aioamqp basic_publish_with_instana error: {exc}")

    @wrapt.patch_function_wrapper("aioamqp.channel", "Channel.basic_consume")
    async def basic_consume_with_instana(
        wrapped: Callable[..., aioamqp.connect],
        instance: object,
        argv: Tuple[object, Tuple[object, ...]],
        kwargs: Dict[str, Any],
    ) -> object:
        if tracing_is_off():
            return await wrapped(*argv, **kwargs)

        callback = argv[0]
        tracer, parent_span, _ = get_tracer_tuple()
        parent_context = parent_span.get_span_context() if parent_span else None

        @wrapt.decorator
        async def callback_wrapper(
            wrapped_callback: Callable[..., aioamqp.connect],
            instance: Any,
            args: Tuple,
            kwargs: Dict,
        ) -> object:
            with tracer.start_as_current_span(
                "aioamqp-consumer", span_context=parent_context
            ) as span:
                try:
                    span.set_status(StatusCode.OK)
                    span.set_attribute("aioamqp.callback", callback)
                    span.set_attribute("aioamqp.message", args[1])
                    span.set_attribute("aioamqp.exchange_name", args[2].exchange_name)
                    span.set_attribute("aioamqp.routing_key", args[2].routing_key)
                    return await wrapped_callback(*args, **kwargs)
                except Exception as exc:
                    span.record_exception(exc)
                    logger.debug(f"aioamqp basic_consume_with_instana error: {exc}")

        wrapped_callback = callback_wrapper(callback)
        argv = (wrapped_callback,) + argv[1:]

        return await wrapped(*argv, **kwargs)

    logger.debug("Instrumenting aioamqp")

except ImportError:
    pass



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

try:
    from typing import TYPE_CHECKING, Any, Callable, Dict, Sequence, Type

    from instana.span_context import SpanContext

    if TYPE_CHECKING:
        from botocore.client import BaseClient
    import wrapt

    from instana.log import logger
    from instana.singletons import tracer
    from instana.util.traceutils import (
        get_tracer_tuple,
        tracing_is_off,
    )

    operations = {
        "upload_file": "UploadFile",
        "upload_fileobj": "UploadFileObj",
        "download_file": "DownloadFile",
        "download_fileobj": "DownloadFileObj",
    }

    def create_s3_span(
        wrapped: Callable[..., Dict[str, Any]],
        instance: Type["BaseClient"],
        args: Sequence[Dict[str, Any]],
        kwargs: Dict[str, Any],
        parent_context: SpanContext,
    ) -> None:
        with tracer.start_as_current_span("s3", span_context=parent_context) as span:
            try:
                span.set_attribute("s3.op", args[0])
                if "Bucket" in args[1].keys():
                    span.set_attribute("s3.bucket", args[1]["Bucket"])
            except Exception as exc:
                span.record_exception(exc)
                logger.debug("create_s3_span: collect error", exc_info=True)

    def collect_s3_injected_attributes(
        wrapped: Callable[..., object],
        instance: Type["BaseClient"],
        args: Sequence[object],
        kwargs: Dict[str, Any],
    ) -> Callable[..., object]:
        # If we're not tracing, just return
        if tracing_is_off():
            return wrapped(*args, **kwargs)

        tracer, parent_span, _ = get_tracer_tuple()

        parent_context = parent_span.get_span_context() if parent_span else None

        with tracer.start_as_current_span("s3", span_context=parent_context) as span:
            try:
                span.set_attribute("s3.op", operations[wrapped.__name__])
                if wrapped.__name__ in ["download_file", "download_fileobj"]:
                    span.set_attribute("s3.bucket", args[0])
                else:
                    span.set_attribute("s3.bucket", args[1])
                return wrapped(*args, **kwargs)
            except Exception as exc:
                span.record_exception(exc)
                logger.debug(
                    "collect_s3_injected_attributes: collect error", exc_info=True
                )

    for method in [
        "upload_file",
        "upload_fileobj",
        "download_file",
        "download_fileobj",
    ]:
        wrapt.wrap_function_wrapper(
            "boto3.s3.inject", method, collect_s3_injected_attributes
        )

    logger.debug("Instrumenting s3")
except ImportError:
    pass



# (c) Copyright IBM Corp. 2025

try:
    import inspect
    from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Tuple

    import kafka  # noqa: F401
    import wrapt
    from opentelemetry.trace import SpanKind

    from instana.log import logger
    from instana.propagators.format import Format
    from instana.util.traceutils import (
        get_tracer_tuple,
        tracing_is_off,
    )

    if TYPE_CHECKING:
        from kafka.producer.future import FutureRecordMetadata

    @wrapt.patch_function_wrapper("kafka", "KafkaProducer.send")
    def trace_kafka_send(
        wrapped: Callable[..., "kafka.KafkaProducer.send"],
        instance: "kafka.KafkaProducer",
        args: Tuple[int, str, Tuple[Any, ...]],
        kwargs: Dict[str, Any],
    ) -> "FutureRecordMetadata":
        if tracing_is_off():
            return wrapped(*args, **kwargs)

        tracer, parent_span, _ = get_tracer_tuple()
        parent_context = parent_span.get_span_context() if parent_span else None

        with tracer.start_as_current_span(
            "kafka-producer", span_context=parent_context, kind=SpanKind.PRODUCER
        ) as span:
            span.set_attribute("kafka.service", args[0])
            span.set_attribute("kafka.access", "send")

            # context propagation
            headers = kwargs.get("headers", [])
            tracer.inject(
                span.context,
                Format.KAFKA_HEADERS,
                headers,
                disable_w3c_trace_context=True,
            )

            try:
                kwargs["headers"] = headers
                res = wrapped(*args, **kwargs)
            except Exception as exc:
                span.record_exception(exc)
            else:
                return res

    def create_span(
        span_type: str,
        topic: Optional[str],
        headers: Optional[List[Tuple[str, bytes]]] = [],
        exception: Optional[str] = None,
    ) -> None:
        tracer, parent_span, _ = get_tracer_tuple()
        parent_context = (
            parent_span.get_span_context()
            if parent_span
            else tracer.extract(
                Format.KAFKA_HEADERS,
                headers,
                disable_w3c_trace_context=True,
            )
        )
        with tracer.start_as_current_span(
            "kafka-consumer", span_context=parent_context, kind=SpanKind.CONSUMER
        ) as span:
            if topic:
                span.set_attribute("kafka.service", topic)
            span.set_attribute("kafka.access", span_type)
            if exception:
                span.record_exception(exception)

    @wrapt.patch_function_wrapper("kafka", "KafkaConsumer.__next__")
    def trace_kafka_consume(
        wrapped: Callable[..., "kafka.KafkaConsumer.__next__"],
        instance: "kafka.KafkaConsumer",
        args: Tuple[int, str, Tuple[Any, ...]],
        kwargs: Dict[str, Any],
    ) -> "FutureRecordMetadata":
        if tracing_is_off():
            return wrapped(*args, **kwargs)

        exception = None
        res = None

        try:
            res = wrapped(*args, **kwargs)
        except Exception as exc:
            exception = exc
        finally:
            if res:
                create_span(
                    "consume",
                    res.topic if res else list(instance.subscription())[0],
                    res.headers,
                )
            else:
                create_span(
                    "consume", list(instance.subscription())[0], exception=exception
                )

        return res

    @wrapt.patch_function_wrapper("kafka", "KafkaConsumer.poll")
    def trace_kafka_poll(
        wrapped: Callable[..., "kafka.KafkaConsumer.poll"],
        instance: "kafka.KafkaConsumer",
        args: Tuple[int, str, Tuple[Any, ...]],
        kwargs: Dict[str, Any],
    ) -> Optional[Dict[str, Any]]:
        if tracing_is_off():
            return wrapped(*args, **kwargs)

        # The KafkaConsumer.consume() from the kafka-python-ng call the
        # KafkaConsumer.poll() internally, so we do not consider it here.
        if any(
            frame.function == "trace_kafka_consume"
            for frame in inspect.getouterframes(inspect.currentframe(), 2)
        ):
            return wrapped(*args, **kwargs)

        exception = None
        res = None

        try:
            res = wrapped(*args, **kwargs)
        except Exception as exc:
            exception = exc
        finally:
            if res:
                for partition, consumer_records in res.items():
                    for message in consumer_records:
                        create_span(
                            "poll",
                            partition.topic,
                            message.headers if hasattr(message, "headers") else [],
                        )
            else:
                create_span(
                    "poll", list(instance.subscription())[0], exception=exception
                )

        return res

    logger.debug("Instrumenting Kafka (kafka-python)")
except ImportError:
    pass



import logging
import os
from typing import Generator

import pytest

from instana.configurator import config
from instana.options import (
    AWSFargateOptions,
    AWSLambdaOptions,
    BaseOptions,
    EKSFargateOptions,
    GCROptions,
    ServerlessOptions,
    StandardOptions,
)

env_vars = [
    "INSTANA_DEBUG",
    "INSTANA_EXTRA_HTTP_HEADERS",
    "INSTANA_IGNORE_ENDPOINTS",
    "INSTANA_SECRETS",
    "INSTANA_AGENT_KEY",
    "INSTANA_ENDPOINT_URL",
    "INSTANA_DISABLE_CA_CHECK",
    "INSTANA_ENDPOINT_PROXY",
    "INSTANA_TIMEOUT",
    "INSTANA_LOG_LEVEL",
]


def clean_env_vars():
    for env_var in env_vars:
        if env_var in os.environ.keys():
            del os.environ[env_var]


class TestBaseOptions:
    @pytest.fixture(autouse=True)
    def _resource(self) -> Generator[None, None, None]:
        yield
        clean_env_vars()
        if "tracing" in config.keys():
            del config["tracing"]

    def test_base_options(self) -> None:
        if "INSTANA_DEBUG" in os.environ:
            del os.environ["INSTANA_DEBUG"]
        test_base_options = BaseOptions()

        assert not test_base_options.debug
        assert test_base_options.log_level == logging.WARN
        assert not test_base_options.extra_http_headers
        assert not test_base_options.allow_exit_as_root
        assert not test_base_options.ignore_endpoints
        assert test_base_options.secrets_matcher == "contains-ignore-case"
        assert test_base_options.secrets_list == ["key", "pass", "secret"]
        assert not test_base_options.secrets

    def test_base_options_with_config(self) -> None:
        config["tracing"]["ignore_endpoints"] = "service1;service3:endpoint1,endpoint2"
        test_base_options = BaseOptions()
        assert test_base_options.ignore_endpoints == [
            "service1",
            "service3.endpoint1",
            "service3.endpoint2",
        ]

    def test_base_options_with_env_vars(self) -> None:
        os.environ["INSTANA_DEBUG"] = "true"
        os.environ["INSTANA_EXTRA_HTTP_HEADERS"] = "SOMETHING;HERE"
        os.environ["INSTANA_IGNORE_ENDPOINTS"] = "service1;service2:endpoint1,endpoint2"
        os.environ["INSTANA_SECRETS"] = "secret1:username,password"

        test_base_options = BaseOptions()
        assert test_base_options.log_level == logging.DEBUG
        assert test_base_options.debug

        assert test_base_options.extra_http_headers == ["something", "here"]

        assert test_base_options.ignore_endpoints == [
            "service1",
            "service2.endpoint1",
            "service2.endpoint2",
        ]

        assert test_base_options.secrets_matcher == "secret1"
        assert test_base_options.secrets_list == ["username", "password"]


class TestStandardOptions:
    @pytest.fixture(autouse=True)
    def _resource(self) -> Generator[None, None, None]:
        yield
        clean_env_vars()
        if "tracing" in config.keys():
            del config["tracing"]

    def test_standard_options(self) -> None:
        test_standard_options = StandardOptions()

        assert test_standard_options.AGENT_DEFAULT_HOST == "localhost"
        assert test_standard_options.AGENT_DEFAULT_PORT == 42699

    def test_set_secrets(self) -> None:
        test_standard_options = StandardOptions()

        test_secrets = {"matcher": "sample-match", "list": ["sample", "list"]}
        test_standard_options.set_secrets(test_secrets)
        assert test_standard_options.secrets_matcher == "sample-match"
        assert test_standard_options.secrets_list == ["sample", "list"]

    def test_set_extra_headers(self) -> None:
        test_standard_options = StandardOptions()
        test_headers = {"header1": "sample-match", "header2": ["sample", "list"]}

        test_standard_options.set_extra_headers(test_headers)
        assert test_standard_options.extra_http_headers == test_headers

    def test_set_tracing(self) -> None:
        test_standard_options = StandardOptions()

        test_tracing = {"ignore-endpoints": "service1;service2:endpoint1,endpoint2"}
        test_standard_options.set_tracing(test_tracing)

        assert test_standard_options.ignore_endpoints == [
            "service1",
            "service2.endpoint1",
            "service2.endpoint2",
        ]
        assert not test_standard_options.extra_http_headers

    def test_set_tracing_priority(self) -> None:
        # Environment variables > In-code Configuration > Agent Configuration
        # First test when all attributes given
        os.environ["INSTANA_IGNORE_ENDPOINTS"] = (
            "env_service1;env_service2:endpoint1,endpoint2"
        )
        config["tracing"]["ignore_endpoints"] = (
            "config_service1;config_service2:endpoint1,endpoint2"
        )
        test_tracing = {"ignore-endpoints": "service1;service2:endpoint1,endpoint2"}

        test_standard_options = StandardOptions()
        test_standard_options.set_tracing(test_tracing)

        assert test_standard_options.ignore_endpoints == [
            "env_service1",
            "env_service2.endpoint1",
            "env_service2.endpoint2",
        ]

        # Second test when In-code configuration and Agent configuration given

        del os.environ["INSTANA_IGNORE_ENDPOINTS"]

        test_standard_options = StandardOptions()
        test_standard_options.set_tracing(test_tracing)

        assert test_standard_options.ignore_endpoints == [
            "config_service1",
            "config_service2.endpoint1",
            "config_service2.endpoint2",
        ]

    def test_set_from(self) -> None:
        test_standard_options = StandardOptions()
        test_res_data = {
            "secrets": {"matcher": "sample-match", "list": ["sample", "list"]},
            "tracing": {"ignore-endpoints": "service1;service2:endpoint1,endpoint2"},
        }
        test_standard_options.set_from(test_res_data)

        assert (
            test_standard_options.secrets_matcher == test_res_data["secrets"]["matcher"]
        )
        assert test_standard_options.secrets_list == test_res_data["secrets"]["list"]
        assert test_standard_options.ignore_endpoints == [
            "service1",
            "service2.endpoint1",
            "service2.endpoint2",
        ]

        test_res_data = {
            "extraHeaders": {"header1": "sample-match", "header2": ["sample", "list"]},
        }
        test_standard_options.set_from(test_res_data)

        assert test_standard_options.extra_http_headers == test_res_data["extraHeaders"]

    def test_set_from_bool(
        self,
        caplog: pytest.LogCaptureFixture,
    ) -> None:
        caplog.set_level(logging.DEBUG, logger="instana")
        caplog.clear()

        test_standard_options = StandardOptions()
        test_res_data = True
        test_standard_options.set_from(test_res_data)

        assert len(caplog.messages) == 1
        assert len(caplog.records) == 1
        assert (
            "options.set_from: Wrong data type - <class 'bool'>" in caplog.messages[0]
        )

        assert test_standard_options.secrets_list == ["key", "pass", "secret"]
        assert test_standard_options.ignore_endpoints == []
        assert not test_standard_options.extra_http_headers


class TestServerlessOptions:
    @pytest.fixture(autouse=True)
    def _resource(self) -> Generator[None, None, None]:
        yield
        clean_env_vars()

    def test_serverless_options(self) -> None:
        test_serverless_options = ServerlessOptions()

        assert not test_serverless_options.debug
        assert test_serverless_options.log_level == logging.WARN
        assert not test_serverless_options.extra_http_headers
        assert not test_serverless_options.allow_exit_as_root
        assert not test_serverless_options.ignore_endpoints
        assert test_serverless_options.secrets_matcher == "contains-ignore-case"
        assert test_serverless_options.secrets_list == ["key", "pass", "secret"]
        assert not test_serverless_options.secrets
        assert not test_serverless_options.agent_key
        assert not test_serverless_options.endpoint_url
        assert test_serverless_options.ssl_verify
        assert not test_serverless_options.endpoint_proxy
        assert test_serverless_options.timeout == 0.8

    def test_serverless_options_with_env_vars(self) -> None:
        os.environ["INSTANA_AGENT_KEY"] = "key1"
        os.environ["INSTANA_ENDPOINT_URL"] = "localhost"
        os.environ["INSTANA_DISABLE_CA_CHECK"] = "true"
        os.environ["INSTANA_ENDPOINT_PROXY"] = "proxy1"
        os.environ["INSTANA_TIMEOUT"] = "3000"
        os.environ["INSTANA_LOG_LEVEL"] = "info"

        test_serverless_options = ServerlessOptions()

        assert test_serverless_options.agent_key == "key1"
        assert test_serverless_options.endpoint_url == "localhost"
        assert not test_serverless_options.ssl_verify
        assert test_serverless_options.endpoint_proxy == {"https": "proxy1"}
        assert test_serverless_options.timeout == 3
        assert test_serverless_options.log_level == logging.INFO


class TestAWSLambdaOptions:
    @pytest.fixture(autouse=True)
    def _resource(self) -> Generator[None, None, None]:
        yield
        clean_env_vars()

    def test_aws_lambda_options(self) -> None:
        test_aws_lambda_options = AWSLambdaOptions()

        assert not test_aws_lambda_options.agent_key
        assert not test_aws_lambda_options.endpoint_url
        assert test_aws_lambda_options.ssl_verify
        assert not test_aws_lambda_options.endpoint_proxy
        assert test_aws_lambda_options.timeout == 0.8
        assert test_aws_lambda_options.log_level == logging.WARN


class TestAWSFargateOptions:
    @pytest.fixture(autouse=True)
    def _resource(self) -> Generator[None, None, None]:
        yield
        clean_env_vars()

    def test_aws_fargate_options(self) -> None:
        test_aws_fargate_options = AWSFargateOptions()

        assert not test_aws_fargate_options.agent_key
        assert not test_aws_fargate_options.endpoint_url
        assert test_aws_fargate_options.ssl_verify
        assert not test_aws_fargate_options.endpoint_proxy
        assert test_aws_fargate_options.timeout == 0.8
        assert test_aws_fargate_options.log_level == logging.WARN
        assert not test_aws_fargate_options.tags
        assert not test_aws_fargate_options.zone

    def test_aws_fargate_options_with_env_vars(self) -> None:
        os.environ["INSTANA_AGENT_KEY"] = "key1"
        os.environ["INSTANA_ENDPOINT_URL"] = "localhost"
        os.environ["INSTANA_DISABLE_CA_CHECK"] = "true"
        os.environ["INSTANA_ENDPOINT_PROXY"] = "proxy1"
        os.environ["INSTANA_TIMEOUT"] = "3000"
        os.environ["INSTANA_LOG_LEVEL"] = "info"
        os.environ["INSTANA_TAGS"] = "key1=value1,key2=value2"
        os.environ["INSTANA_ZONE"] = "zone1"

        test_aws_fargate_options = AWSFargateOptions()

        assert test_aws_fargate_options.agent_key == "key1"
        assert test_aws_fargate_options.endpoint_url == "localhost"
        assert not test_aws_fargate_options.ssl_verify
        assert test_aws_fargate_options.endpoint_proxy == {"https": "proxy1"}
        assert test_aws_fargate_options.timeout == 3
        assert test_aws_fargate_options.log_level == logging.INFO

        assert test_aws_fargate_options.tags == {"key1": "value1", "key2": "value2"}
        assert test_aws_fargate_options.zone == "zone1"


class TestEKSFargateOptions:
    @pytest.fixture(autouse=True)
    def _resource(self) -> Generator[None, None, None]:
        yield
        clean_env_vars()

    def test_eks_fargate_options(self) -> None:
        test_eks_fargate_options = EKSFargateOptions()

        assert not test_eks_fargate_options.agent_key
        assert not test_eks_fargate_options.endpoint_url
        assert test_eks_fargate_options.ssl_verify
        assert not test_eks_fargate_options.endpoint_proxy
        assert test_eks_fargate_options.timeout == 0.8
        assert test_eks_fargate_options.log_level == logging.WARN

    def test_eks_fargate_options_with_env_vars(self) -> None:
        os.environ["INSTANA_AGENT_KEY"] = "key1"
        os.environ["INSTANA_ENDPOINT_URL"] = "localhost"
        os.environ["INSTANA_DISABLE_CA_CHECK"] = "true"
        os.environ["INSTANA_ENDPOINT_PROXY"] = "proxy1"
        os.environ["INSTANA_TIMEOUT"] = "3000"
        os.environ["INSTANA_LOG_LEVEL"] = "info"

        test_eks_fargate_options = EKSFargateOptions()

        assert test_eks_fargate_options.agent_key == "key1"
        assert test_eks_fargate_options.endpoint_url == "localhost"
        assert not test_eks_fargate_options.ssl_verify
        assert test_eks_fargate_options.endpoint_proxy == {"https": "proxy1"}
        assert test_eks_fargate_options.timeout == 3
        assert test_eks_fargate_options.log_level == logging.INFO


class TestGCROptions:
    @pytest.fixture(autouse=True)
    def _resource(self) -> Generator[None, None, None]:
        yield
        clean_env_vars()

    def test_gcr_options(self) -> None:
        test_gcr_options = GCROptions()

        assert not test_gcr_options.debug
        assert test_gcr_options.log_level == logging.WARN
        assert not test_gcr_options.extra_http_headers
        assert not test_gcr_options.allow_exit_as_root
        assert not test_gcr_options.ignore_endpoints
        assert test_gcr_options.secrets_matcher == "contains-ignore-case"
        assert test_gcr_options.secrets_list == ["key", "pass", "secret"]
        assert not test_gcr_options.secrets
        assert not test_gcr_options.agent_key
        assert not test_gcr_options.endpoint_url
        assert test_gcr_options.ssl_verify
        assert not test_gcr_options.endpoint_proxy
        assert test_gcr_options.timeout == 0.8

    def test_gcr_options_with_env_vars(self) -> None:
        os.environ["INSTANA_AGENT_KEY"] = "key1"
        os.environ["INSTANA_ENDPOINT_URL"] = "localhost"
        os.environ["INSTANA_DISABLE_CA_CHECK"] = "true"
        os.environ["INSTANA_ENDPOINT_PROXY"] = "proxy1"
        os.environ["INSTANA_TIMEOUT"] = "3000"
        os.environ["INSTANA_LOG_LEVEL"] = "info"

        test_gcr_options = GCROptions()

        assert test_gcr_options.agent_key == "key1"
        assert test_gcr_options.endpoint_url == "localhost"
        assert not test_gcr_options.ssl_verify
        assert test_gcr_options.endpoint_proxy == {"https": "proxy1"}
        assert test_gcr_options.timeout == 3
        assert test_gcr_options.log_level == logging.INFO



# (c) Copyright IBM Corp. 2025

import time
import urllib3
import pytest
from typing import Generator

from tests.apps import spyne_app
from tests.helpers import testenv
from instana.singletons import agent, tracer
from instana.span.span import get_current_span
from instana.util.ids import hex_id


class TestSpyne:
    @pytest.fixture(autouse=True)
    def _resource(self) -> Generator[None, None, None]:
        """Clear all spans before a test run"""
        self.http = urllib3.PoolManager()
        self.recorder = tracer.span_processor
        self.recorder.clear_spans()
        time.sleep(0.1)

    def test_vanilla_requests(self) -> None:
        response = self.http.request("GET", testenv["spyne_server"] + "/hello")
        spans = self.recorder.queued_spans()

        assert len(spans) == 1
        assert get_current_span().is_recording() is False
        assert response.status == 200

    def test_get_request(self) -> None:
        with tracer.start_as_current_span("test"):
            response = self.http.request("GET", testenv["spyne_server"] + "/hello")

        spans = self.recorder.queued_spans()

        assert len(spans) == 3
        assert get_current_span().is_recording() is False

        spyne_span = spans[0]
        urllib3_span = spans[1]
        test_span = spans[2]

        assert response
        assert response.status == 200

        assert "X-INSTANA-T" in response.headers
        assert int(response.headers["X-INSTANA-T"], 16)
        assert response.headers["X-INSTANA-T"] == hex_id(spyne_span.t)

        assert "X-INSTANA-S" in response.headers
        assert int(response.headers["X-INSTANA-S"], 16)
        assert response.headers["X-INSTANA-S"] == hex_id(spyne_span.s)

        assert "X-INSTANA-L" in response.headers
        assert response.headers["X-INSTANA-L"] == "1"

        assert "Server-Timing" in response.headers
        server_timing_value = f"intid;desc={hex_id(spyne_span.t)}"
        assert response.headers["Server-Timing"] == server_timing_value

        # Same traceId
        assert test_span.t == urllib3_span.t
        assert urllib3_span.t == spyne_span.t

        # Parent relationships
        assert urllib3_span.p == test_span.s
        assert spyne_span.p == urllib3_span.s

        assert spyne_span.sy is None
        assert urllib3_span.sy is None
        assert test_span.sy is None

        # Error logging
        assert test_span.ec is None
        assert urllib3_span.ec is None
        assert spyne_span.ec is None

        # spyne
        assert spyne_span.n == "rpc-server"
        assert spyne_span.data["rpc"]["host"] == "127.0.0.1"
        assert spyne_span.data["rpc"]["call"] == "/hello"
        assert spyne_span.data["rpc"]["port"] == str(testenv["spyne_port"])
        assert spyne_span.data["rpc"]["error"] is None
        assert spyne_span.stack is None

    def test_secret_scrubbing(self) -> None:
        with tracer.start_as_current_span("test"):
            response = self.http.request("GET", testenv["spyne_server"] + "/say_hello?name=World&times=4&secret=sshhh")

        spans = self.recorder.queued_spans()

        assert len(spans) == 3
        assert get_current_span().is_recording() is False

        spyne_span = spans[0]
        urllib3_span = spans[1]
        test_span = spans[2]

        assert response
        assert response.status == 200

        assert "X-INSTANA-T" in response.headers
        assert int(response.headers["X-INSTANA-T"], 16)
        assert response.headers["X-INSTANA-T"] == hex_id(spyne_span.t)

        assert "X-INSTANA-S" in response.headers
        assert int(response.headers["X-INSTANA-S"], 16)
        assert response.headers["X-INSTANA-S"] == hex_id(spyne_span.s)

        assert "X-INSTANA-L" in response.headers
        assert response.headers["X-INSTANA-L"] == "1"

        assert "Server-Timing" in response.headers
        server_timing_value = f"intid;desc={hex_id(spyne_span.t)}"
        assert response.headers["Server-Timing"] == server_timing_value

        # Same traceId
        assert test_span.t == urllib3_span.t
        assert urllib3_span.t == spyne_span.t

        # Parent relationships
        assert urllib3_span.p == test_span.s
        assert spyne_span.p == urllib3_span.s

        assert spyne_span.sy is None
        assert urllib3_span.sy is None
        assert test_span.sy is None

        # Error logging
        assert test_span.ec is None
        assert urllib3_span.ec is None
        assert spyne_span.ec is None

        # spyne
        assert spyne_span.n == "rpc-server"
        assert spyne_span.data["rpc"]["host"] == "127.0.0.1"
        assert spyne_span.data["rpc"]["call"] == "/say_hello"
        assert spyne_span.data["rpc"]["params"] == "name=World&times=4&secret=<redacted>"
        assert spyne_span.data["rpc"]["port"] == str(testenv["spyne_port"])
        assert spyne_span.data["rpc"]["error"] is None
        assert spyne_span.stack is None

    def test_custom_404(self) -> None:
        with tracer.start_as_current_span("test"):
            response = self.http.request("GET", testenv["spyne_server"] + "/custom_404?user_id=9876")

        spans = self.recorder.queued_spans()

        assert len(spans) == 4
        assert get_current_span().is_recording() is False

        log_span = spans[0]
        spyne_span = spans[1]
        urllib3_span = spans[2]
        test_span = spans[3]

        assert response
        assert response.status == 404

        assert "X-INSTANA-T" in response.headers
        assert int(response.headers["X-INSTANA-T"], 16)
        assert response.headers["X-INSTANA-T"] == hex_id(spyne_span.t)

        assert "X-INSTANA-S" in response.headers
        assert int(response.headers["X-INSTANA-S"], 16)
        assert response.headers["X-INSTANA-S"] == hex_id(spyne_span.s)

        assert "X-INSTANA-L" in response.headers
        assert response.headers["X-INSTANA-L"] == "1"

        assert "Server-Timing" in response.headers
        server_timing_value = f"intid;desc={hex_id(spyne_span.t)}"
        assert response.headers["Server-Timing"] == server_timing_value

        # Same traceId
        assert test_span.t == urllib3_span.t
        assert urllib3_span.t == spyne_span.t

        # Parent relationships
        assert urllib3_span.p == test_span.s
        assert spyne_span.p == urllib3_span.s

        # Synthetic
        assert spyne_span.sy is None
        assert urllib3_span.sy is None
        assert test_span.sy is None

        # Error logging
        assert test_span.ec is None
        assert urllib3_span.ec is None
        assert spyne_span.ec is None

        # spyne
        assert spyne_span.n == "rpc-server"
        assert spyne_span.data["rpc"]["host"] == "127.0.0.1"
        assert spyne_span.data["rpc"]["call"] == "/custom_404"
        assert spyne_span.data["rpc"]["port"] == str(testenv["spyne_port"])
        assert spyne_span.data["rpc"]["params"] == "user_id=9876"
        assert spyne_span.data["rpc"]["error"] is None
        assert spyne_span.stack is None

        # urllib3
        assert test_span.data["sdk"]["name"] == "test"
        assert urllib3_span.n == "urllib3"
        assert urllib3_span.data["http"]["status"] == 404
        assert (
            testenv["spyne_server"] + "/custom_404" == urllib3_span.data["http"]["url"]
        )
        assert urllib3_span.data["http"]["method"] == "GET"
        assert urllib3_span.stack is not None
        assert type(urllib3_span.stack) is list
        assert len(urllib3_span.stack) > 1

    def test_404(self) -> None:
        with tracer.start_as_current_span("test"):
            response = self.http.request("GET", testenv["spyne_server"] + "/11111")

        spans = self.recorder.queued_spans()

        assert len(spans) == 3
        assert get_current_span().is_recording() is False

        spyne_span = spans[0]
        urllib3_span = spans[1]
        test_span = spans[2]

        assert response
        assert response.status == 404

        assert "X-INSTANA-T" in response.headers
        assert int(response.headers["X-INSTANA-T"], 16)
        assert response.headers["X-INSTANA-T"] == hex_id(spyne_span.t)

        assert "X-INSTANA-S" in response.headers
        assert int(response.headers["X-INSTANA-S"], 16)
        assert response.headers["X-INSTANA-S"] == hex_id(spyne_span.s)

        assert "X-INSTANA-L" in response.headers
        assert response.headers["X-INSTANA-L"] == "1"

        assert "Server-Timing" in response.headers
        server_timing_value = f"intid;desc={hex_id(spyne_span.t)}"
        assert response.headers["Server-Timing"] == server_timing_value

        # Same traceId
        assert test_span.t == urllib3_span.t
        assert urllib3_span.t == spyne_span.t

        # Parent relationships
        assert urllib3_span.p == test_span.s
        assert spyne_span.p == urllib3_span.s

        # Synthetic
        assert spyne_span.sy is None
        assert urllib3_span.sy is None
        assert test_span.sy is None

        # Error logging
        assert test_span.ec is None
        assert urllib3_span.ec is None
        assert spyne_span.ec is None

        # spyne
        assert spyne_span.n == "rpc-server"
        assert spyne_span.data["rpc"]["host"] == "127.0.0.1"
        assert spyne_span.data["rpc"]["call"] == "/11111"
        assert spyne_span.data["rpc"]["port"] == str(testenv["spyne_port"])
        assert spyne_span.data["rpc"]["error"] is None
        assert spyne_span.stack is None

        # urllib3
        assert test_span.data["sdk"]["name"] == "test"
        assert urllib3_span.n == "urllib3"
        assert urllib3_span.data["http"]["status"] == 404
        assert (
            testenv["spyne_server"] + "/11111" == urllib3_span.data["http"]["url"]
        )
        assert urllib3_span.data["http"]["method"] == "GET"
        assert urllib3_span.stack is not None
        assert type(urllib3_span.stack) is list
        assert len(urllib3_span.stack) > 1

    def test_500(self) -> None:
        with tracer.start_as_current_span("test"):
            response = self.http.request("GET", testenv["spyne_server"] + "/exception")

        spans = self.recorder.queued_spans()

        assert len(spans) == 4
        assert get_current_span().is_recording() is False

        log_span = spans[0]
        spyne_span = spans[1]
        urllib3_span = spans[2]
        test_span = spans[3]

        assert response
        assert response.status == 500

        assert "X-INSTANA-T" in response.headers
        assert int(response.headers["X-INSTANA-T"], 16)
        assert response.headers["X-INSTANA-T"] == hex_id(spyne_span.t)

        assert "X-INSTANA-S" in response.headers
        assert int(response.headers["X-INSTANA-S"], 16)
        assert response.headers["X-INSTANA-S"] == hex_id(spyne_span.s)

        assert "X-INSTANA-L" in response.headers
        assert response.headers["X-INSTANA-L"] == "1"

        assert "Server-Timing" in response.headers
        server_timing_value = f"intid;desc={hex_id(spyne_span.t)}"
        assert response.headers["Server-Timing"] == server_timing_value

        # Same traceId
        assert test_span.t == urllib3_span.t
        assert urllib3_span.t == spyne_span.t

        # Parent relationships
        assert urllib3_span.p == test_span.s
        assert spyne_span.p == urllib3_span.s

        assert spyne_span.sy is None
        assert urllib3_span.sy is None
        assert test_span.sy is None

        # Error logging
        assert test_span.ec is None
        assert urllib3_span.ec == 1
        assert spyne_span.ec == 1

        # spyne
        assert spyne_span.n == "rpc-server"
        assert spyne_span.data["rpc"]["host"] == "127.0.0.1"
        assert spyne_span.data["rpc"]["call"] == "/exception"
        assert spyne_span.data["rpc"]["port"] == str(testenv["spyne_port"])        
        assert spyne_span.data["rpc"]["error"]
        assert spyne_span.stack is None



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

import os
from typing import Generator

import boto3
import pytest
from moto import mock_aws

from instana.options import StandardOptions
from instana.singletons import agent, tracer
from tests.helpers import get_first_span_by_filter


class TestDynamoDB:
    @pytest.fixture(autouse=True)
    def _resource(self) -> Generator[None, None, None]:
        self.recorder = tracer.span_processor
        self.recorder.clear_spans()
        self.mock = mock_aws()
        self.mock.start()
        self.dynamodb = boto3.client("dynamodb", region_name="us-west-2")
        yield
        self.mock.stop()
        agent.options.allow_exit_as_root = False

    def test_vanilla_create_table(self) -> None:
        self.dynamodb.create_table(
            TableName="dynamodb-table",
            KeySchema=[{"AttributeName": "id", "KeyType": "HASH"}],
            AttributeDefinitions=[{"AttributeName": "id", "AttributeType": "S"}],
            ProvisionedThroughput={"ReadCapacityUnits": 1, "WriteCapacityUnits": 1},
        )
        result = self.dynamodb.list_tables()
        assert len(result["TableNames"]) == 1
        assert result["TableNames"][0] == "dynamodb-table"

    def test_dynamodb_create_table(self) -> None:
        with tracer.start_as_current_span("test"):
            self.dynamodb.create_table(
                TableName="dynamodb-table",
                KeySchema=[{"AttributeName": "id", "KeyType": "HASH"}],
                AttributeDefinitions=[{"AttributeName": "id", "AttributeType": "S"}],
                ProvisionedThroughput={"ReadCapacityUnits": 1, "WriteCapacityUnits": 1},
            )
        result = self.dynamodb.list_tables()
        assert len(result["TableNames"]) == 1
        assert result["TableNames"][0] == "dynamodb-table"

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        filter = lambda span: span.n == "sdk"  # noqa: E731
        test_span = get_first_span_by_filter(spans, filter)
        assert test_span

        filter = lambda span: span.n == "dynamodb"  # noqa: E731
        dynamodb_span = get_first_span_by_filter(spans, filter)
        assert dynamodb_span

        assert dynamodb_span.t == test_span.t
        assert dynamodb_span.p == test_span.s

        assert not test_span.ec
        assert not dynamodb_span.ec

        assert dynamodb_span.data["dynamodb"]["op"] == "CreateTable"
        assert dynamodb_span.data["dynamodb"]["region"] == "us-west-2"
        assert dynamodb_span.data["dynamodb"]["table"] == "dynamodb-table"

    def test_ignore_dynamodb(self) -> None:
        os.environ["INSTANA_IGNORE_ENDPOINTS"] = "dynamodb"
        agent.options = StandardOptions()

        with tracer.start_as_current_span("test"):
            self.dynamodb.create_table(
                TableName="dynamodb-table",
                KeySchema=[{"AttributeName": "id", "KeyType": "HASH"}],
                AttributeDefinitions=[{"AttributeName": "id", "AttributeType": "S"}],
                ProvisionedThroughput={"ReadCapacityUnits": 1, "WriteCapacityUnits": 1},
            )

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        filter = lambda span: span.n == "dynamodb"  # noqa: E731
        dynamodb_span = get_first_span_by_filter(spans, filter)
        assert dynamodb_span

        filtered_spans = agent.filter_spans(spans)
        assert len(filtered_spans) == 1

        assert dynamodb_span not in filtered_spans

    def test_ignore_create_table(self) -> None:
        os.environ["INSTANA_IGNORE_ENDPOINTS"] = "dynamodb.createtable"
        agent.options = StandardOptions()

        with tracer.start_as_current_span("test"):
            self.dynamodb.create_table(
                TableName="dynamodb-table",
                KeySchema=[{"AttributeName": "id", "KeyType": "HASH"}],
                AttributeDefinitions=[{"AttributeName": "id", "AttributeType": "S"}],
                ProvisionedThroughput={"ReadCapacityUnits": 1, "WriteCapacityUnits": 1},
            )
            self.dynamodb.list_tables()

        spans = self.recorder.queued_spans()
        assert len(spans) == 3

        filtered_spans = agent.filter_spans(spans)
        assert len(filtered_spans) == 2

        filter = lambda span: span.n == "dynamodb"  # noqa: E731
        dynamodb_span = get_first_span_by_filter(filtered_spans, filter)

        assert dynamodb_span.n == "dynamodb"
        assert dynamodb_span.data["dynamodb"]["op"] == "ListTables"

    def test_dynamodb_create_table_as_root_exit_span(self) -> None:
        agent.options.allow_exit_as_root = True
        self.dynamodb.create_table(
            TableName="dynamodb-table",
            KeySchema=[{"AttributeName": "id", "KeyType": "HASH"}],
            AttributeDefinitions=[{"AttributeName": "id", "AttributeType": "S"}],
            ProvisionedThroughput={"ReadCapacityUnits": 1, "WriteCapacityUnits": 1},
        )
        agent.options.allow_exit_as_root = False
        result = self.dynamodb.list_tables()
        assert len(result["TableNames"]) == 1
        assert result["TableNames"][0] == "dynamodb-table"

        spans = self.recorder.queued_spans()
        assert len(spans) == 1
        dynamodb_span = spans[0]
        assert dynamodb_span
        assert dynamodb_span.n == "dynamodb"
        assert not dynamodb_span.p
        assert not dynamodb_span.ec

        assert dynamodb_span.data["dynamodb"]["op"] == "CreateTable"
        assert dynamodb_span.data["dynamodb"]["region"] == "us-west-2"
        assert dynamodb_span.data["dynamodb"]["table"] == "dynamodb-table"

    def test_dynamodb_list_tables(self) -> None:
        with tracer.start_as_current_span("test"):
            result = self.dynamodb.list_tables()

        assert len(result["TableNames"]) == 0
        assert result["ResponseMetadata"]["HTTPStatusCode"] == 200

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        filter = lambda span: span.n == "sdk"  # noqa: E731
        test_span = get_first_span_by_filter(spans, filter)
        assert test_span

        filter = lambda span: span.n == "dynamodb"  # noqa: E731
        dynamodb_span = get_first_span_by_filter(spans, filter)
        assert dynamodb_span

        assert dynamodb_span.t == test_span.t
        assert dynamodb_span.p == test_span.s

        assert not test_span.ec
        assert not dynamodb_span.ec

        assert dynamodb_span.data["dynamodb"]["op"] == "ListTables"
        assert dynamodb_span.data["dynamodb"]["region"] == "us-west-2"

    def test_dynamodb_put_item(self) -> None:
        self.dynamodb.create_table(
            TableName="dynamodb-table",
            KeySchema=[{"AttributeName": "id", "KeyType": "HASH"}],
            AttributeDefinitions=[{"AttributeName": "id", "AttributeType": "S"}],
            ProvisionedThroughput={"ReadCapacityUnits": 1, "WriteCapacityUnits": 1},
        )
        with tracer.start_as_current_span("test"):
            self.dynamodb.put_item(
                TableName="dynamodb-table",
                Item={"id": {"S": "1"}, "name": {"S": "John"}},
            )

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        filter = lambda span: span.n == "sdk"  # noqa: E731
        test_span = get_first_span_by_filter(spans, filter)
        assert test_span

        filter = lambda span: span.n == "dynamodb"  # noqa: E731
        dynamodb_span = get_first_span_by_filter(spans, filter)
        assert dynamodb_span

        assert dynamodb_span.t == test_span.t
        assert dynamodb_span.p == test_span.s

        assert not test_span.ec
        assert not dynamodb_span.ec

        assert dynamodb_span.data["dynamodb"]["op"] == "PutItem"
        assert dynamodb_span.data["dynamodb"]["region"] == "us-west-2"
        assert dynamodb_span.data["dynamodb"]["table"] == "dynamodb-table"

    def test_dynamodb_scan(self) -> None:
        test_item = {"id": {"S": "1"}, "name": {"S": "John"}}
        self.dynamodb.create_table(
            TableName="dynamodb-table",
            KeySchema=[{"AttributeName": "id", "KeyType": "HASH"}],
            AttributeDefinitions=[{"AttributeName": "id", "AttributeType": "S"}],
            ProvisionedThroughput={"ReadCapacityUnits": 1, "WriteCapacityUnits": 1},
        )
        self.dynamodb.put_item(
            TableName="dynamodb-table",
            Item=test_item,
        )
        with tracer.start_as_current_span("test"):
            result = self.dynamodb.scan(TableName="dynamodb-table")

        assert result["Items"] == [test_item]
        assert result["Count"] == 1
        assert result["ScannedCount"] == 1
        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        filter = lambda span: span.n == "sdk"  # noqa: E731
        test_span = get_first_span_by_filter(spans, filter)
        assert test_span

        filter = lambda span: span.n == "dynamodb"  # noqa: E731
        dynamodb_span = get_first_span_by_filter(spans, filter)
        assert dynamodb_span

        assert dynamodb_span.t == test_span.t
        assert dynamodb_span.p == test_span.s

        assert not test_span.ec
        assert not dynamodb_span.ec

        assert dynamodb_span.data["dynamodb"]["op"] == "Scan"
        assert dynamodb_span.data["dynamodb"]["region"] == "us-west-2"
        assert dynamodb_span.data["dynamodb"]["table"] == "dynamodb-table"

    def test_dynamodb_get_item(self) -> None:
        test_item = {"id": {"S": "1"}, "name": {"S": "John"}}
        self.dynamodb.create_table(
            TableName="dynamodb-table",
            KeySchema=[{"AttributeName": "id", "KeyType": "HASH"}],
            AttributeDefinitions=[{"AttributeName": "id", "AttributeType": "S"}],
            ProvisionedThroughput={"ReadCapacityUnits": 1, "WriteCapacityUnits": 1},
        )
        self.dynamodb.put_item(
            TableName="dynamodb-table",
            Item=test_item,
        )
        with tracer.start_as_current_span("test"):
            result = self.dynamodb.get_item(
                TableName="dynamodb-table", Key={"id": {"S": "1"}}
            )

        assert result["Item"] == test_item
        assert result["ResponseMetadata"]

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        filter = lambda span: span.n == "sdk"  # noqa: E731
        test_span = get_first_span_by_filter(spans, filter)
        assert test_span

        filter = lambda span: span.n == "dynamodb"  # noqa: E731
        dynamodb_span = get_first_span_by_filter(spans, filter)
        assert dynamodb_span

        assert dynamodb_span.t == test_span.t
        assert dynamodb_span.p == test_span.s

        assert not test_span.ec
        assert not dynamodb_span.ec

        assert dynamodb_span.data["dynamodb"]["op"] == "GetItem"
        assert dynamodb_span.data["dynamodb"]["region"] == "us-west-2"
        assert dynamodb_span.data["dynamodb"]["table"] == "dynamodb-table"

    def test_dynamodb_update_item(self) -> None:
        test_item = {"id": {"S": "1"}, "name": {"S": "John"}}
        self.dynamodb.create_table(
            TableName="dynamodb-table",
            KeySchema=[{"AttributeName": "id", "KeyType": "HASH"}],
            AttributeDefinitions=[{"AttributeName": "id", "AttributeType": "S"}],
            ProvisionedThroughput={"ReadCapacityUnits": 1, "WriteCapacityUnits": 1},
        )
        self.dynamodb.put_item(
            TableName="dynamodb-table",
            Item=test_item,
        )
        with tracer.start_as_current_span("test"):
            self.dynamodb.update_item(
                TableName="dynamodb-table",
                Key={"id": {"S": "1"}},  # Specify the key
                UpdateExpression="SET #attr_name = :new_name",
                ExpressionAttributeNames={"#attr_name": "name"},  # Use alias for "name"
                ExpressionAttributeValues={":new_name": {"S": "Updated John"}},
                ReturnValues="UPDATED_NEW",
            )

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        filter = lambda span: span.n == "sdk"  # noqa: E731
        test_span = get_first_span_by_filter(spans, filter)
        assert test_span

        filter = lambda span: span.n == "dynamodb"  # noqa: E731
        dynamodb_span = get_first_span_by_filter(spans, filter)
        assert dynamodb_span

        assert dynamodb_span.t == test_span.t
        assert dynamodb_span.p == test_span.s

        assert not test_span.ec
        assert not dynamodb_span.ec

        assert dynamodb_span.data["dynamodb"]["op"] == "UpdateItem"
        assert dynamodb_span.data["dynamodb"]["region"] == "us-west-2"
        assert dynamodb_span.data["dynamodb"]["table"] == "dynamodb-table"

    def test_dynamodb_delete_item(self) -> None:
        test_item = {"id": {"S": "1"}, "name": {"S": "John"}}
        self.dynamodb.create_table(
            TableName="dynamodb-table",
            KeySchema=[{"AttributeName": "id", "KeyType": "HASH"}],
            AttributeDefinitions=[{"AttributeName": "id", "AttributeType": "S"}],
            ProvisionedThroughput={"ReadCapacityUnits": 1, "WriteCapacityUnits": 1},
        )
        self.dynamodb.put_item(
            TableName="dynamodb-table",
            Item=test_item,
        )
        with tracer.start_as_current_span("test"):
            self.dynamodb.delete_item(
                TableName="dynamodb-table", Key={"id": {"S": "1"}}
            )

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        filter = lambda span: span.n == "sdk"  # noqa: E731
        test_span = get_first_span_by_filter(spans, filter)
        assert test_span

        filter = lambda span: span.n == "dynamodb"  # noqa: E731
        dynamodb_span = get_first_span_by_filter(spans, filter)
        assert dynamodb_span

        assert dynamodb_span.t == test_span.t
        assert dynamodb_span.p == test_span.s

        assert not test_span.ec
        assert not dynamodb_span.ec

        assert dynamodb_span.data["dynamodb"]["op"] == "DeleteItem"
        assert dynamodb_span.data["dynamodb"]["region"] == "us-west-2"
        assert dynamodb_span.data["dynamodb"]["table"] == "dynamodb-table"

    def test_dynamodb_query_item(self) -> None:
        test_item = {"id": {"S": "1"}, "name": {"S": "John"}}
        self.dynamodb.create_table(
            TableName="dynamodb-table",
            KeySchema=[{"AttributeName": "id", "KeyType": "HASH"}],
            AttributeDefinitions=[{"AttributeName": "id", "AttributeType": "S"}],
            ProvisionedThroughput={"ReadCapacityUnits": 1, "WriteCapacityUnits": 1},
        )
        self.dynamodb.put_item(
            TableName="dynamodb-table",
            Item=test_item,
        )
        self.dynamodb.put_item(
            TableName="dynamodb-table", Item={"id": {"S": "2"}, "name": {"S": "Jack"}}
        )
        with tracer.start_as_current_span("test"):
            self.dynamodb.query(
                TableName="dynamodb-table",
                KeyConditionExpression="id = :pk_val",
                ExpressionAttributeValues={":pk_val": {"S": "1"}},
            )

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        filter = lambda span: span.n == "sdk"  # noqa: E731
        test_span = get_first_span_by_filter(spans, filter)
        assert test_span

        filter = lambda span: span.n == "dynamodb"  # noqa: E731
        dynamodb_span = get_first_span_by_filter(spans, filter)
        assert dynamodb_span

        assert dynamodb_span.t == test_span.t
        assert dynamodb_span.p == test_span.s

        assert not test_span.ec
        assert not dynamodb_span.ec

        assert dynamodb_span.data["dynamodb"]["op"] == "Query"
        assert dynamodb_span.data["dynamodb"]["region"] == "us-west-2"
        assert dynamodb_span.data["dynamodb"]["table"] == "dynamodb-table"



# (c) Copyright IBM Corp. 2025

try:
    from typing import Any, Callable, Dict, List, Optional, Tuple

    import confluent_kafka  # noqa: F401
    import wrapt
    from confluent_kafka import Consumer, Producer
    from opentelemetry.trace import SpanKind

    from instana.log import logger
    from instana.propagators.format import Format
    from instana.util.traceutils import (
        get_tracer_tuple,
        tracing_is_off,
    )

    # As confluent_kafka is a wrapper around the C-developed librdkafka
    # (provided automatically via binary wheels), we have to create new classes
    # inheriting from the confluent_kafka package with the methods to be
    # monkey-patched.
    class InstanaConfluentKafkaProducer(Producer):
        """
        Wrapper class for confluent_kafka.Producer, which is an Asynchronous Kafka Producer.
        """

        def produce(
            self,
            topic: str,
            *args: object,
            **kwargs: Dict[str, Any],
        ) -> None:
            return super().produce(topic, *args, **kwargs)

    class InstanaConfluentKafkaConsumer(Consumer):
        """
        Wrapper class for confluent_kafka.Consumer, which is a high-level Apache Kafka consumer.
        """

        def consume(
            self, *args: object, **kwargs: Dict[str, Any]
        ) -> List[confluent_kafka.Message]:
            return super().consume(*args, **kwargs)

        def poll(
            self, timeout: Optional[float] = -1
        ) -> Optional[confluent_kafka.Message]:
            return super().poll(timeout)

    def trace_kafka_produce(
        wrapped: Callable[..., InstanaConfluentKafkaProducer.produce],
        instance: InstanaConfluentKafkaProducer,
        args: Tuple[int, str, Tuple[Any, ...]],
        kwargs: Dict[str, Any],
    ) -> None:
        if tracing_is_off():
            return wrapped(*args, **kwargs)

        tracer, parent_span, _ = get_tracer_tuple()
        parent_context = parent_span.get_span_context() if parent_span else None

        with tracer.start_as_current_span(
            "kafka-producer", span_context=parent_context, kind=SpanKind.PRODUCER
        ) as span:
            span.set_attribute("kafka.service", args[0])
            span.set_attribute("kafka.access", "produce")

            # context propagation
            #
            # As stated in the official documentation at
            # https://docs.confluent.io/platform/current/clients/confluent-kafka-python/html/index.html#pythonclient-producer,
            # headers can be either a list of (key, value) pairs or a
            # dictionary. To maintain compatibility with the headers for the
            # Kafka Python library, we will use a list of tuples.
            headers = args[6] if len(args) > 6 else kwargs.get("headers", [])
            tracer.inject(
                span.context,
                Format.KAFKA_HEADERS,
                headers,
                disable_w3c_trace_context=True,
            )

            try:
                kwargs["headers"] = headers
                res = wrapped(*args, **kwargs)
            except Exception as exc:
                span.record_exception(exc)
            else:
                return res

    def create_span(
        span_type: str,
        topic: Optional[str] = "",
        headers: Optional[List[Tuple[str, bytes]]] = [],
        exception: Optional[str] = None,
    ) -> None:
        tracer, parent_span, _ = get_tracer_tuple()
        parent_context = (
            parent_span.get_span_context()
            if parent_span
            else tracer.extract(
                Format.KAFKA_HEADERS,
                headers,
                disable_w3c_trace_context=True,
            )
        )
        with tracer.start_as_current_span(
            "kafka-consumer", span_context=parent_context, kind=SpanKind.CONSUMER
        ) as span:
            if topic:
                span.set_attribute("kafka.service", topic)
            span.set_attribute("kafka.access", span_type)

            if exception:
                span.record_exception(exception)

    def trace_kafka_consume(
        wrapped: Callable[..., InstanaConfluentKafkaConsumer.consume],
        instance: InstanaConfluentKafkaConsumer,
        args: Tuple[int, str, Tuple[Any, ...]],
        kwargs: Dict[str, Any],
    ) -> List[confluent_kafka.Message]:
        if tracing_is_off():
            return wrapped(*args, **kwargs)

        res = None
        exception = None

        try:
            res = wrapped(*args, **kwargs)
        except Exception as exc:
            exception = exc
        finally:
            if res:
                for message in res:
                    create_span("consume", message.topic(), message.headers())
            else:
                create_span("consume", exception=exception)

        return res

    def trace_kafka_poll(
        wrapped: Callable[..., InstanaConfluentKafkaConsumer.poll],
        instance: InstanaConfluentKafkaConsumer,
        args: Tuple[int, str, Tuple[Any, ...]],
        kwargs: Dict[str, Any],
    ) -> Optional[confluent_kafka.Message]:
        if tracing_is_off():
            return wrapped(*args, **kwargs)

        res = None
        exception = None

        try:
            res = wrapped(*args, **kwargs)
        except Exception as exc:
            exception = exc
        finally:
            if res:
                create_span("poll", res.topic(), res.headers())
            else:
                create_span(
                    "poll",
                    next(iter(instance.list_topics().topics)),
                    exception=exception,
                )

        return res

    # Apply the monkey patch
    confluent_kafka.Producer = InstanaConfluentKafkaProducer
    confluent_kafka.Consumer = InstanaConfluentKafkaConsumer

    wrapt.wrap_function_wrapper(
        InstanaConfluentKafkaProducer, "produce", trace_kafka_produce
    )
    wrapt.wrap_function_wrapper(
        InstanaConfluentKafkaConsumer, "consume", trace_kafka_consume
    )
    wrapt.wrap_function_wrapper(InstanaConfluentKafkaConsumer, "poll", trace_kafka_poll)

    logger.debug("Instrumenting Kafka (confluent_kafka)")
except ImportError:
    pass



# (c) Copyright IBM Corp. 2025

from typing import Generator

import pytest
from confluent_kafka import (
    Consumer,
    KafkaException,
    Producer,
)
from confluent_kafka.admin import AdminClient, NewTopic
from opentelemetry.trace import SpanKind

from instana.singletons import agent, tracer
from tests.helpers import get_first_span_by_filter, testenv


class TestConfluentKafka:
    @pytest.fixture(autouse=True)
    def _resource(self) -> Generator[None, None, None]:
        """SetUp and TearDown"""
        # setup
        # Clear all spans before a test run
        self.recorder = tracer.span_processor
        self.recorder.clear_spans()

        # Kafka admin client
        self.kafka_config = {"bootstrap.servers": testenv["kafka_bootstrap_servers"][0]}
        self.kafka_client = AdminClient(self.kafka_config)

        try:
            topics = self.kafka_client.create_topics(  # noqa: F841
                [
                    NewTopic(
                        testenv["kafka_topic"],
                        num_partitions=1,
                        replication_factor=1,
                    ),
                ]
            )
        except KafkaException:
            pass

        # Kafka producer
        self.producer = Producer(self.kafka_config)
        yield
        # teardown
        # Ensure that allow_exit_as_root has the default value"""
        agent.options.allow_exit_as_root = False
        # Close connections
        self.kafka_client.delete_topics([testenv["kafka_topic"]])

    def test_trace_confluent_kafka_produce(self) -> None:
        with tracer.start_as_current_span("test"):
            self.producer.produce(testenv["kafka_topic"], b"raw_bytes")
            self.producer.flush(timeout=10)

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        kafka_span = spans[0]
        test_span = spans[1]

        # Same traceId
        assert test_span.t == kafka_span.t

        # Parent relationships
        assert kafka_span.p == test_span.s

        # Error logging
        assert not test_span.ec
        assert not kafka_span.ec

        assert kafka_span.n == "kafka"
        assert kafka_span.k == SpanKind.CLIENT
        assert kafka_span.data["kafka"]["service"] == testenv["kafka_topic"]
        assert kafka_span.data["kafka"]["access"] == "produce"

    def test_trace_confluent_kafka_consume(self) -> None:
        # Produce some events
        self.producer.produce(testenv["kafka_topic"], value=b"raw_bytes1")
        self.producer.flush(timeout=30)

        # Consume the events
        consumer_config = self.kafka_config.copy()
        consumer_config["group.id"] = "my-group"
        consumer_config["auto.offset.reset"] = "earliest"

        consumer = Consumer(consumer_config)
        consumer.subscribe([testenv["kafka_topic"]])

        with tracer.start_as_current_span("test"):
            msgs = consumer.consume(num_messages=1, timeout=60)  # noqa: F841

        consumer.close()

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        kafka_span = spans[0]
        test_span = spans[1]

        # Same traceId
        assert test_span.t == kafka_span.t

        # Parent relationships
        assert kafka_span.p == test_span.s

        # Error logging
        assert not test_span.ec
        assert not kafka_span.ec

        assert kafka_span.n == "kafka"
        assert kafka_span.k == SpanKind.SERVER
        assert kafka_span.data["kafka"]["access"] == "consume"

    def test_trace_confluent_kafka_poll(self) -> None:
        # Produce some events
        self.producer.produce(testenv["kafka_topic"], b"raw_bytes1")
        self.producer.flush()

        # Consume the events
        consumer_config = self.kafka_config.copy()
        consumer_config["group.id"] = "my-group"
        consumer_config["auto.offset.reset"] = "earliest"

        consumer = Consumer(consumer_config)
        consumer.subscribe([testenv["kafka_topic"]])

        with tracer.start_as_current_span("test"):
            msg = consumer.poll(timeout=30)  # noqa: F841

        consumer.close()

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        kafka_span = spans[0]
        test_span = spans[1]

        # Same traceId
        assert test_span.t == kafka_span.t

        # Parent relationships
        assert kafka_span.p == test_span.s

        assert kafka_span.n == "kafka"
        assert kafka_span.k == SpanKind.SERVER
        assert kafka_span.data["kafka"]["access"] == "poll"

    def test_trace_confluent_kafka_error(self) -> None:
        # Consume the events
        consumer_config = {"bootstrap.servers": ["some_inexistent_host:9094"]}
        consumer_config["group.id"] = "my-group"
        consumer_config["auto.offset.reset"] = "earliest"

        consumer = Consumer(consumer_config)
        consumer.subscribe(["inexistent_kafka_topic"])

        with tracer.start_as_current_span("test"):
            consumer.consume(-10)

        consumer.close()

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        kafka_span = spans[0]
        test_span = spans[len(spans) - 1]

        # Same traceId
        assert test_span.t == kafka_span.t

        # Parent relationships
        assert kafka_span.p == test_span.s

        # Error logging
        assert not test_span.ec
        assert kafka_span.ec == 1

        assert kafka_span.n == "kafka"
        assert kafka_span.k == SpanKind.SERVER
        assert not kafka_span.data["kafka"]["service"]
        assert kafka_span.data["kafka"]["access"] == "consume"
        assert (
            kafka_span.data["kafka"]["error"]
            == "num_messages must be between 0 and 1000000 (1M)"
        )

    def test_confluent_kafka_consumer_root_exit(self) -> None:
        agent.options.allow_exit_as_root = True

        self.producer.produce(testenv["kafka_topic"] + "_1", b"raw_bytes")
        self.producer.produce(testenv["kafka_topic"] + "_2", b"raw_bytes")
        self.producer.flush(timeout=10)

        # Consume the events
        consumer_config = self.kafka_config.copy()
        consumer_config["group.id"] = "my-group"
        consumer_config["auto.offset.reset"] = "earliest"

        consumer = Consumer(consumer_config)
        consumer.subscribe(
            [
                testenv["kafka_topic"] + "_1",
                testenv["kafka_topic"] + "_2",
            ]
        )

        consumer.consume(num_messages=2, timeout=60)  # noqa: F841

        consumer.close()

        spans = self.recorder.queued_spans()
        assert len(spans) == 4

        producer_span_1 = get_first_span_by_filter(
            spans,
            lambda span: span.n == "kafka"
            and span.data["kafka"]["access"] == "produce"
            and span.data["kafka"]["service"] == "span-topic_1",
        )
        producer_span_2 = get_first_span_by_filter(
            spans,
            lambda span: span.n == "kafka"
            and span.data["kafka"]["access"] == "produce"
            and span.data["kafka"]["service"] == "span-topic_2",
        )
        consumer_span_1 = get_first_span_by_filter(
            spans,
            lambda span: span.n == "kafka"
            and span.data["kafka"]["access"] == "consume"
            and span.data["kafka"]["service"] == "span-topic_1",
        )
        consumer_span_2 = get_first_span_by_filter(
            spans,
            lambda span: span.n == "kafka"
            and span.data["kafka"]["access"] == "consume"
            and span.data["kafka"]["service"] == "span-topic_2",
        )

        # same trace id, different span ids
        assert producer_span_1.t == consumer_span_1.t
        assert producer_span_1.s == consumer_span_1.p
        assert producer_span_1.s != consumer_span_1.s

        assert producer_span_2.t == consumer_span_2.t
        assert producer_span_2.s == consumer_span_2.p
        assert producer_span_2.s != consumer_span_2.s

        self.kafka_client.delete_topics(
            [
                testenv["kafka_topic"] + "_1",
                testenv["kafka_topic"] + "_2",
            ]
        )

    def test_confluent_kafka_poll_root_exit(self) -> None:
        agent.options.allow_exit_as_root = True

        # Produce some events
        self.producer.produce(testenv["kafka_topic"], b"raw_bytes1")
        self.producer.flush()

        # Consume the events
        consumer_config = self.kafka_config.copy()
        consumer_config["group.id"] = "my-group"
        consumer_config["auto.offset.reset"] = "earliest"

        consumer = Consumer(consumer_config)
        consumer.subscribe([testenv["kafka_topic"]])

        msg = consumer.poll(timeout=30)  # noqa: F841

        consumer.close()

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        producer_span = get_first_span_by_filter(
            spans,
            lambda span: span.n == "kafka"
            and span.data["kafka"]["access"] == "produce"
            and span.data["kafka"]["service"] == "span-topic",
        )

        poll_span = get_first_span_by_filter(
            spans,
            lambda span: span.n == "kafka"
            and span.data["kafka"]["access"] == "poll"
            and span.data["kafka"]["service"] == "span-topic",
        )

        # Same traceId
        assert producer_span.t == poll_span.t
        assert producer_span.s == poll_span.p
        assert producer_span.s != poll_span.s

    def test_confluent_kafka_poll_root_exit_error(self) -> None:
        agent.options.allow_exit_as_root = True

        # Produce some events
        self.producer.produce(testenv["kafka_topic"], b"raw_bytes1")
        self.producer.flush()

        # Consume the events
        consumer_config = self.kafka_config.copy()
        consumer_config["group.id"] = "my-group"
        consumer_config["auto.offset.reset"] = "earliest"

        consumer = Consumer(consumer_config)
        consumer.subscribe([testenv["kafka_topic"]])

        msg = consumer.poll(timeout="wrong_value")  # noqa: F841

        consumer.close()

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        poll_span = get_first_span_by_filter(
            spans,
            lambda span: span.n == "kafka"
            and span.data["kafka"]["access"] == "poll"
            and span.data["kafka"]["service"] == "span-topic",
        )
        assert poll_span.data["kafka"]["error"] == "must be real number, not str"



from typing import Optional
import pytest

from instana.util.span_utils import get_operation_specifier


@pytest.mark.parametrize(
    "span_name, expected_result",
    [("something", ""), ("redis", "command"), ("dynamodb", "op")],
)
def test_get_operation_specifier(
    span_name: str, expected_result: Optional[str]
) -> None:
    response_redis = get_operation_specifier(span_name)
    assert response_redis == expected_result



# (c) Copyright IBM Corp. 2025

from typing import Generator

import pytest
from kafka import KafkaConsumer, KafkaProducer
from kafka.admin import KafkaAdminClient, NewTopic
from kafka.errors import TopicAlreadyExistsError
from opentelemetry.trace import SpanKind

from instana.singletons import agent, tracer
from tests.helpers import get_first_span_by_filter, testenv


class TestKafkaPython:
    @pytest.fixture(autouse=True)
    def _resource(self) -> Generator[None, None, None]:
        """SetUp and TearDown"""
        # setup
        # Clear all spans before a test run
        self.recorder = tracer.span_processor
        self.recorder.clear_spans()

        # Kafka admin client
        self.kafka_client = KafkaAdminClient(
            bootstrap_servers=testenv["kafka_bootstrap_servers"],
            client_id="test_kafka_python",
        )

        try:
            self.kafka_client.create_topics(
                [
                    NewTopic(
                        name=testenv["kafka_topic"],
                        num_partitions=1,
                        replication_factor=1,
                    ),
                ]
            )
        except TopicAlreadyExistsError:
            pass

        # Kafka producer
        self.producer = KafkaProducer(
            bootstrap_servers=testenv["kafka_bootstrap_servers"]
        )
        yield
        # teardown
        # Ensure that allow_exit_as_root has the default value"""
        agent.options.allow_exit_as_root = False
        # Close connections
        self.producer.close()
        self.kafka_client.delete_topics([testenv["kafka_topic"]])
        self.kafka_client.close()

    def test_trace_kafka_python_send(self) -> None:
        with tracer.start_as_current_span("test"):
            future = self.producer.send(testenv["kafka_topic"], b"raw_bytes")

        record_metadata = future.get(timeout=10)  # noqa: F841

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        kafka_span = spans[0]
        test_span = spans[1]

        # Same traceId
        assert test_span.t == kafka_span.t

        # Parent relationships
        assert kafka_span.p == test_span.s

        # Error logging
        assert not test_span.ec
        assert not kafka_span.ec

        assert kafka_span.n == "kafka"
        assert kafka_span.k == SpanKind.CLIENT
        assert kafka_span.data["kafka"]["service"] == testenv["kafka_topic"]
        assert kafka_span.data["kafka"]["access"] == "send"

    def test_trace_kafka_python_consume(self) -> None:
        # Produce some events
        self.producer.send(testenv["kafka_topic"], b"raw_bytes1")
        self.producer.send(testenv["kafka_topic"], b"raw_bytes2")
        self.producer.flush()

        # Consume the events
        consumer = KafkaConsumer(
            testenv["kafka_topic"],
            bootstrap_servers=testenv["kafka_bootstrap_servers"],
            auto_offset_reset="earliest",  # consume earliest available messages
            enable_auto_commit=False,  # do not auto-commit offsets
            consumer_timeout_ms=1000,
        )

        with tracer.start_as_current_span("test"):
            for msg in consumer:
                if msg is None:
                    break

        consumer.close()

        spans = self.recorder.queued_spans()
        assert len(spans) == 4

        kafka_span = spans[0]
        test_span = spans[len(spans) - 1]

        # Same traceId
        assert test_span.t == kafka_span.t

        # Parent relationships
        assert kafka_span.p == test_span.s

        # Error logging
        assert not test_span.ec
        assert not kafka_span.ec

        assert kafka_span.n == "kafka"
        assert kafka_span.k == SpanKind.SERVER
        assert kafka_span.data["kafka"]["service"] == testenv["kafka_topic"]
        assert kafka_span.data["kafka"]["access"] == "consume"

    def test_trace_kafka_python_poll(self) -> None:
        # Produce some events
        self.producer.send(testenv["kafka_topic"], b"raw_bytes1")
        self.producer.send(testenv["kafka_topic"], b"raw_bytes2")
        self.producer.flush()

        # Consume the events
        consumer = KafkaConsumer(
            testenv["kafka_topic"],
            bootstrap_servers=testenv["kafka_bootstrap_servers"],
            auto_offset_reset="earliest",  # consume earliest available messages
            enable_auto_commit=False,  # do not auto-commit offsets
            consumer_timeout_ms=1000,
        )

        with tracer.start_as_current_span("test"):
            msg = consumer.poll()  # noqa: F841

        consumer.close()

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        kafka_span = spans[0]
        test_span = spans[1]

        # Same traceId
        assert test_span.t == kafka_span.t

        # Parent relationships
        assert kafka_span.p == test_span.s

        # Error logging
        assert not test_span.ec
        assert not kafka_span.ec

        assert kafka_span.n == "kafka"
        assert kafka_span.k == SpanKind.SERVER
        assert kafka_span.data["kafka"]["service"] == testenv["kafka_topic"]
        assert kafka_span.data["kafka"]["access"] == "poll"

    def test_trace_kafka_python_error(self) -> None:
        # Consume the events
        consumer = KafkaConsumer(
            "inexistent_kafka_topic",
            bootstrap_servers=testenv["kafka_bootstrap_servers"],
            auto_offset_reset="earliest",  # consume earliest available messages
            enable_auto_commit=False,  # do not auto-commit offsets
            consumer_timeout_ms=1000,
        )

        with tracer.start_as_current_span("test"):
            for msg in consumer:
                if msg is None:
                    break

        consumer.close()

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        kafka_span = spans[0]
        test_span = spans[1]

        # Same traceId
        assert test_span.t == kafka_span.t

        # Parent relationships
        assert kafka_span.p == test_span.s

        # Error logging
        assert not test_span.ec
        assert kafka_span.ec == 1

        assert kafka_span.n == "kafka"
        assert kafka_span.k == SpanKind.SERVER
        assert kafka_span.data["kafka"]["service"] == "inexistent_kafka_topic"
        assert kafka_span.data["kafka"]["access"] == "consume"
        assert kafka_span.data["kafka"]["error"] == "StopIteration()"

    def test_kafka_consumer_root_exit(self) -> None:
        agent.options.allow_exit_as_root = True

        self.producer.send(testenv["kafka_topic"], b"raw_bytes")
        self.producer.flush()

        # Consume the events
        consumer = KafkaConsumer(
            testenv["kafka_topic"],
            bootstrap_servers=testenv["kafka_bootstrap_servers"],
            auto_offset_reset="earliest",  # consume earliest available messages
            enable_auto_commit=False,  # do not auto-commit offsets
            consumer_timeout_ms=1000,
        )

        for msg in consumer:
            if msg is None:
                break

        consumer.close()

        spans = self.recorder.queued_spans()
        assert len(spans) == 4

        producer_span = spans[0]
        consumer_span = spans[1]

        assert producer_span.s
        assert producer_span.n == "kafka"
        assert producer_span.data["kafka"]["access"] == "send"
        assert producer_span.data["kafka"]["service"] == "span-topic"

        assert consumer_span.s
        assert consumer_span.n == "kafka"
        assert consumer_span.data["kafka"]["access"] == "consume"
        assert consumer_span.data["kafka"]["service"] == "span-topic"

        assert producer_span.t == consumer_span.t

    def test_kafka_poll_root_exit(self) -> None:
        agent.options.allow_exit_as_root = True

        self.kafka_client.create_topics(
            [
                NewTopic(
                    name=testenv["kafka_topic"] + "_1",
                    num_partitions=1,
                    replication_factor=1,
                ),
                NewTopic(
                    name=testenv["kafka_topic"] + "_2",
                    num_partitions=1,
                    replication_factor=1,
                ),
                NewTopic(
                    name=testenv["kafka_topic"] + "_3",
                    num_partitions=1,
                    replication_factor=1,
                ),
            ]
        )

        self.producer.send(testenv["kafka_topic"] + "_1", b"raw_bytes1")
        self.producer.send(testenv["kafka_topic"] + "_2", b"raw_bytes2")
        self.producer.send(testenv["kafka_topic"] + "_3", b"raw_bytes3")
        self.producer.flush()

        # Consume the events
        consumer = KafkaConsumer(
            bootstrap_servers=testenv["kafka_bootstrap_servers"],
            auto_offset_reset="earliest",  # consume earliest available messages
            enable_auto_commit=False,  # do not auto-commit offsets
            consumer_timeout_ms=1000,
        )
        topics = [
            testenv["kafka_topic"] + "_1",
            testenv["kafka_topic"] + "_2",
            testenv["kafka_topic"] + "_3",
        ]
        consumer.subscribe(topics)

        messages = consumer.poll(timeout_ms=1000)  # noqa: F841
        consumer.close()

        spans = self.recorder.queued_spans()
        assert len(spans) == 6

        producer_span_1 = get_first_span_by_filter(
            spans,
            lambda span: span.n == "kafka"
            and span.data["kafka"]["access"] == "send"
            and span.data["kafka"]["service"] == "span-topic_1",
        )
        producer_span_2 = get_first_span_by_filter(
            spans,
            lambda span: span.n == "kafka"
            and span.data["kafka"]["access"] == "send"
            and span.data["kafka"]["service"] == "span-topic_2",
        )
        producer_span_3 = get_first_span_by_filter(
            spans,
            lambda span: span.n == "kafka"
            and span.data["kafka"]["access"] == "send"
            and span.data["kafka"]["service"] == "span-topic_3",
        )

        poll_span_1 = get_first_span_by_filter(
            spans,
            lambda span: span.n == "kafka"
            and span.data["kafka"]["access"] == "poll"
            and span.data["kafka"]["service"] == "span-topic_1",
        )
        poll_span_2 = get_first_span_by_filter(
            spans,
            lambda span: span.n == "kafka"
            and span.data["kafka"]["access"] == "poll"
            and span.data["kafka"]["service"] == "span-topic_2",
        )
        poll_span_3 = get_first_span_by_filter(
            spans,
            lambda span: span.n == "kafka"
            and span.data["kafka"]["access"] == "poll"
            and span.data["kafka"]["service"] == "span-topic_3",
        )

        assert producer_span_1.n == "kafka"
        assert producer_span_1.data["kafka"]["access"] == "send"
        assert producer_span_1.data["kafka"]["service"] == "span-topic_1"

        assert producer_span_2.n == "kafka"
        assert producer_span_2.data["kafka"]["access"] == "send"
        assert producer_span_2.data["kafka"]["service"] == "span-topic_2"

        assert producer_span_3.n == "kafka"
        assert producer_span_3.data["kafka"]["access"] == "send"
        assert producer_span_3.data["kafka"]["service"] == "span-topic_3"

        assert poll_span_1.n == "kafka"
        assert poll_span_1.data["kafka"]["access"] == "poll"
        assert poll_span_1.data["kafka"]["service"] == "span-topic_1"

        assert poll_span_2.n == "kafka"
        assert poll_span_2.data["kafka"]["access"] == "poll"
        assert poll_span_2.data["kafka"]["service"] == "span-topic_2"

        assert poll_span_3.n == "kafka"
        assert poll_span_3.data["kafka"]["access"] == "poll"
        assert poll_span_3.data["kafka"]["service"] == "span-topic_3"

        # same trace id, different span ids
        assert producer_span_1.t == poll_span_1.t
        assert producer_span_1.s != poll_span_1.s

        assert producer_span_2.t == poll_span_2.t
        assert producer_span_2.s != poll_span_2.s

        assert producer_span_3.t == poll_span_3.t
        assert producer_span_3.s != poll_span_3.s

        self.kafka_client.delete_topics(
            [
                testenv["kafka_topic"] + "_1",
                testenv["kafka_topic"] + "_2",
                testenv["kafka_topic"] + "_3",
            ]
        )



# (c) Copyright IBM Corp. 2025

from typing import Optional


def get_operation_specifier(span_name: str) -> Optional[str]:
    """Get the specific operation specifier for the given span."""
    operation_specifier = ""
    if span_name == "redis":
        operation_specifier = "command"
    elif span_name == "dynamodb":
        operation_specifier = "op"
    return operation_specifier



# (c) Copyright IBM Corp. 2025
try:
    from typing import TYPE_CHECKING, Any, Callable, Dict, Sequence, Tuple, Type

    from opentelemetry.semconv.trace import SpanAttributes

    from instana.instrumentation.aws.dynamodb import create_dynamodb_span
    from instana.instrumentation.aws.s3 import create_s3_span

    if TYPE_CHECKING:
        from botocore.auth import SigV4Auth
        from botocore.client import BaseClient

        from instana.span.span import InstanaSpan

    import json

    import wrapt

    from instana.log import logger
    from instana.propagators.format import Format
    from instana.singletons import tracer
    from instana.span.span import get_current_span
    from instana.util.traceutils import (
        extract_custom_headers,
        get_tracer_tuple,
        tracing_is_off,
    )

    def lambda_inject_context(payload: Dict[str, Any], span: "InstanaSpan") -> None:
        """
        When boto3 lambda client 'Invoke' is called, we want to inject the tracing context.
        boto3/botocore has specific requirements:
        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/lambda.html#Lambda.Client.invoke
        """
        try:
            invoke_payload = payload.get("Payload", {})

            if not isinstance(invoke_payload, dict):
                invoke_payload = json.loads(invoke_payload)

            tracer.inject(span.context, Format.HTTP_HEADERS, invoke_payload)
            payload["Payload"] = json.dumps(invoke_payload)
        except Exception:
            logger.debug("non-fatal lambda_inject_context: ", exc_info=True)

    @wrapt.patch_function_wrapper("botocore.auth", "SigV4Auth.add_auth")
    def emit_add_auth_with_instana(
        wrapped: Callable[..., None],
        instance: "SigV4Auth",
        args: Tuple[object],
        kwargs: Dict[str, Any],
    ) -> Callable[..., None]:
        current_span = get_current_span()
        if not tracing_is_off() and current_span and current_span.is_recording():
            extract_custom_headers(current_span, args[0].headers)
        return wrapped(*args, **kwargs)

    @wrapt.patch_function_wrapper("botocore.client", "BaseClient._make_api_call")
    def make_api_call_with_instana(
        wrapped: Callable[..., Dict[str, Any]],
        instance: Type["BaseClient"],
        args: Sequence[Dict[str, Any]],
        kwargs: Dict[str, Any],
    ) -> Dict[str, Any]:
        # If we're not tracing, just return
        if tracing_is_off():
            return wrapped(*args, **kwargs)

        tracer, parent_span, _ = get_tracer_tuple()

        parent_context = parent_span.get_span_context() if parent_span else None

        if instance.meta.service_model.service_name == "dynamodb":
            create_dynamodb_span(wrapped, instance, args, kwargs, parent_context)
        elif instance.meta.service_model.service_name == "s3":
            create_s3_span(wrapped, instance, args, kwargs, parent_context)
        else:
            with tracer.start_as_current_span(
                "boto3", span_context=parent_context
            ) as span:
                operation = args[0]
                payload = args[1]

                span.set_attribute("op", operation)
                span.set_attribute("ep", instance._endpoint.host)
                span.set_attribute("reg", instance._client_config.region_name)

                span.set_attribute(
                    SpanAttributes.HTTP_URL,
                    instance._endpoint.host + ":443/" + args[0],
                )
                span.set_attribute(SpanAttributes.HTTP_METHOD, "POST")

                # Don't collect payload for SecretsManager
                if not hasattr(instance, "get_secret_value"):
                    span.set_attribute("payload", payload)

                # Inject context when invoking lambdas
                if "lambda" in instance._endpoint.host and operation == "Invoke":
                    lambda_inject_context(payload, span)

                try:
                    result = wrapped(*args, **kwargs)

                    if isinstance(result, dict):
                        http_dict = result.get("ResponseMetadata")
                        if isinstance(http_dict, dict):
                            status = http_dict.get("HTTPStatusCode")
                            if status is not None:
                                span.set_attribute("http.status_code", status)
                            headers = http_dict.get("HTTPHeaders")
                            extract_custom_headers(span, headers)

                    return result
                except Exception as exc:
                    span.mark_as_errored({"error": exc})
                    raise
        return wrapped(*args, **kwargs)

except ImportError:
    pass



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2016

"""
Option classes for the in-process Instana agent

The description and hierarchy of the classes in this file are as follows:

BaseOptions - base class for all environments.  Holds settings common to all.
  - StandardOptions - The options class used when running directly on a host/node with an Instana agent
  - ServerlessOptions - Base class for serverless environments.  Holds settings common to all serverless environments.
    - AWSLambdaOptions - Options class for AWS Lambda.  Holds settings specific to AWS Lambda.
    - AWSFargateOptions - Options class for AWS Fargate.  Holds settings specific to AWS Fargate.
    - GCROptions - Options class for Google cloud Run.  Holds settings specific to GCR.
"""

import os
import logging
from typing import Any, Dict

from instana.log import logger
from instana.util.config import parse_ignored_endpoints
from instana.util.runtime import determine_service_name
from instana.configurator import config


class BaseOptions(object):
    """Base class for all option classes.  Holds items common to all"""

    def __init__(self, **kwds: Dict[str, Any]) -> None:
        self.debug = False
        self.log_level = logging.WARN
        self.service_name = determine_service_name()
        self.extra_http_headers = None
        self.allow_exit_as_root = False
        self.ignore_endpoints = []

        if "INSTANA_DEBUG" in os.environ:
            self.log_level = logging.DEBUG
            self.debug = True

        if "INSTANA_EXTRA_HTTP_HEADERS" in os.environ:
            self.extra_http_headers = (
                str(os.environ["INSTANA_EXTRA_HTTP_HEADERS"]).lower().split(";")
            )

        if "INSTANA_IGNORE_ENDPOINTS" in os.environ:
            self.ignore_endpoints = parse_ignored_endpoints(
                os.environ["INSTANA_IGNORE_ENDPOINTS"]
            )
        else:
            if (
                isinstance(config.get("tracing"), dict)
                and "ignore_endpoints" in config["tracing"]
            ):
                self.ignore_endpoints = parse_ignored_endpoints(
                    config["tracing"]["ignore_endpoints"],
                )

        if os.environ.get("INSTANA_ALLOW_EXIT_AS_ROOT", None) == "1":
            self.allow_exit_as_root = True

        # Defaults
        self.secrets_matcher = "contains-ignore-case"
        self.secrets_list = ["key", "pass", "secret"]

        # Env var format: <matcher>:<secret>[,<secret>]
        self.secrets = os.environ.get("INSTANA_SECRETS", None)

        if self.secrets is not None:
            parts = self.secrets.split(":")
            if len(parts) == 2:
                self.secrets_matcher = parts[0]
                self.secrets_list = parts[1].split(",")
            else:
                logger.warning(
                    f"Couldn't parse INSTANA_SECRETS env var: {self.secrets}"
                )

        self.__dict__.update(kwds)


class StandardOptions(BaseOptions):
    """The options class used when running directly on a host/node with an Instana agent"""

    AGENT_DEFAULT_HOST = "localhost"
    AGENT_DEFAULT_PORT = 42699

    def __init__(self, **kwds: Dict[str, Any]) -> None:
        super(StandardOptions, self).__init__()

        self.agent_host = os.environ.get("INSTANA_AGENT_HOST", self.AGENT_DEFAULT_HOST)
        self.agent_port = os.environ.get("INSTANA_AGENT_PORT", self.AGENT_DEFAULT_PORT)

        if not isinstance(self.agent_port, int):
            self.agent_port = int(self.agent_port)

    def set_secrets(self, secrets: Dict[str, Any]) -> None:
        """
        Set the secret option from the agent config.
        @param secrets: dictionary of secrets
        @return: None
        """
        self.secrets_matcher = secrets["matcher"]
        self.secrets_list = secrets["list"]

    def set_extra_headers(self, extra_headers: Dict[str, Any]) -> None:
        """
        Set the extra headers option from the agent config, which uses the legacy configuration setting.
        @param extra_headers: dictionary of headers
        @return: None
        """
        if self.extra_http_headers is None:
            self.extra_http_headers = extra_headers
        else:
            self.extra_http_headers.extend(extra_headers)
        logger.info(
            f"Will also capture these custom headers: {self.extra_http_headers}"
        )

    def set_tracing(self, tracing: Dict[str, Any]) -> None:
        """
        Set tracing options from the agent config.
        @param tracing: tracing configuration dictionary
        @return: None
        """
        if (
            "ignore-endpoints" in tracing
            and "INSTANA_IGNORE_ENDPOINTS" not in os.environ
            and "tracing" not in config
        ):
            self.ignore_endpoints = parse_ignored_endpoints(tracing["ignore-endpoints"])
        if "extra-http-headers" in tracing:
            self.extra_http_headers = tracing["extra-http-headers"]

    def set_from(self, res_data: Dict[str, Any]) -> None:
        """
        Set the source identifiers given to use by the Instana Host agent.
        @param res_data: source identifiers provided as announce response
        @return: None
        """
        if not res_data or not isinstance(res_data, dict):
            logger.debug(f"options.set_from: Wrong data type - {type(res_data)}")
            return 

        if "secrets" in res_data:
            self.set_secrets(res_data["secrets"])

        if "tracing" in res_data:
            self.set_tracing(res_data["tracing"])
        else:
            if "extraHeaders" in res_data:
                self.set_extra_headers(res_data["extraHeaders"])


class ServerlessOptions(BaseOptions):
    """Base class for serverless environments.  Holds settings common to all serverless environments."""

    def __init__(self, **kwds: Dict[str, Any]) -> None:
        super(ServerlessOptions, self).__init__()

        self.agent_key = os.environ.get("INSTANA_AGENT_KEY", None)
        self.endpoint_url = os.environ.get("INSTANA_ENDPOINT_URL", None)

        # Remove any trailing slash (if any)
        if self.endpoint_url is not None and self.endpoint_url[-1] == "/":
            self.endpoint_url = self.endpoint_url[:-1]

        if "INSTANA_DISABLE_CA_CHECK" in os.environ:
            self.ssl_verify = False
        else:
            self.ssl_verify = True

        proxy = os.environ.get("INSTANA_ENDPOINT_PROXY", None)
        if proxy is None:
            self.endpoint_proxy = {}
        else:
            self.endpoint_proxy = {"https": proxy}

        timeout_in_ms = os.environ.get("INSTANA_TIMEOUT", None)
        if timeout_in_ms is None:
            self.timeout = 0.8
        else:
            # Convert the value from milliseconds to seconds for the requests package
            try:
                self.timeout = int(timeout_in_ms) / 1000
            except ValueError:
                logger.warning(
                    f"Likely invalid INSTANA_TIMEOUT={timeout_in_ms} value.  Using default."
                )
                logger.warning(
                    "INSTANA_TIMEOUT should specify timeout in milliseconds.  See "
                    "https://www.instana.com/docs/reference/environment_variables/#serverless-monitoring"
                )
                self.timeout = 0.8

        value = os.environ.get("INSTANA_LOG_LEVEL", None)
        if value is not None:
            try:
                value = value.lower()
                if value == "debug":
                    self.log_level = logging.DEBUG
                elif value == "info":
                    self.log_level = logging.INFO
                elif value == "warn" or value == "warning":
                    self.log_level = logging.WARNING
                elif value == "error":
                    self.log_level = logging.ERROR
                else:
                    logger.warning(f"Unknown INSTANA_LOG_LEVEL specified: {value}")
            except Exception:
                logger.debug("BaseAgent.update_log_level: ", exc_info=True)


class AWSLambdaOptions(ServerlessOptions):
    """Options class for AWS Lambda.  Holds settings specific to AWS Lambda."""

    def __init__(self, **kwds: Dict[str, Any]) -> None:
        super(AWSLambdaOptions, self).__init__()


class AWSFargateOptions(ServerlessOptions):
    """Options class for AWS Fargate.  Holds settings specific to AWS Fargate."""

    def __init__(self, **kwds: Dict[str, Any]) -> None:
        super(AWSFargateOptions, self).__init__()

        self.tags = None
        tag_list = os.environ.get("INSTANA_TAGS", None)
        if tag_list is not None:
            try:
                self.tags = dict()
                tags = tag_list.split(",")
                for tag_and_value in tags:
                    parts = tag_and_value.split("=")
                    length = len(parts)
                    if length == 1:
                        self.tags[parts[0]] = None
                    elif length == 2:
                        self.tags[parts[0]] = parts[1]
            except Exception:
                logger.debug(f"Error parsing INSTANA_TAGS env var: {tag_list}")

        self.zone = os.environ.get("INSTANA_ZONE", None)


class EKSFargateOptions(AWSFargateOptions):
    """Options class for EKS Pods on AWS Fargate. Holds settings specific to EKS Pods on AWS Fargate."""

    def __init__(self, **kwds: Dict[str, Any]) -> None:
        super(EKSFargateOptions, self).__init__()


class GCROptions(ServerlessOptions):
    """Options class for Google Cloud Run.  Holds settings specific to Google Cloud Run."""

    def __init__(self, **kwds: Dict[str, Any]) -> None:
        super(GCROptions, self).__init__()
